//
//  Shader_Blitter_Facing_Textured.vsh
//
//  Copyright (c) 2012 Midoki Ltd. All rights reserved.
//

#define SHADER_LIGHT
#define SHADER_LIGHT_CLASSIC
//#define SHADER_LIGHT_SH
#define SHADER_FOG
#define SHADER_FOG_SUN_HAZE

attribute vec3 position;
attribute vec4 texCoord;
attribute vec3 rotScale;
attribute vec4 lightingSettings;
attribute vec4 colour;
attribute vec4 ambientLightColour;

varying lowp vec4 colorVarying;
varying mediump vec4 texCoordHueVarying;

#if defined (SHADER_FOG)
varying mediump	vec4 varying_fogColour;
#endif

uniform mat4 viewWorld;
uniform mat4 viewScreen;

#if defined (SHADER_LIGHT_CLASSIC)
uniform highp	mat4 lightDirectionMat;
uniform highp	mat4 lightColourMat;
#endif

#if defined (SHADER_LIGHT_SH)
uniform mediump	vec4 lightSHCoeff[7];
uniform mediump	vec3 lightDirection;
uniform mediump	vec3 lightColour;
#endif

#if defined (SHADER_FOG)
#define SHADER_FOG_EXP2
uniform mediump	vec4 uniform_fogPlaneY;
uniform mediump	vec4 uniform_fogPlaneZ;
uniform mediump vec4 uniform_fogColour;
#endif

#if defined (SHADER_FOG_SUN_HAZE)
uniform mediump vec4 uniform_sunHazeColourFog;
#endif

//------------------------------------------------------------------------------

#if defined (SHADER_LIGHT_CLASSIC)
mediump vec4 calculateClassicLightingDots(vec4 n, mat4 ldirmat)
{
    mediump vec4	light_dots = (ldirmat * n);
    
//    // Half lambert type wrap...
//    light_dots = ((light_dots * 0.25) + 0.25);

//    light_dots		= max(light_dots, 0.0);
    light_dots		= clamp(light_dots, 0.0, 1.0);
    
    light_dots.w	= 1.0;
    
    return light_dots;
}
#endif

//------------------------------------------------------------------------------

#if defined (SHADER_LIGHT_SH)
mediump vec3 calculateSphericalHarmonics(mediump vec3 n)
{
    //LM: I've removed the commented out code and defines, see gl_uber_main.vsh for the commented version
    mediump vec3	flip	= (n.xzy * n.zyx);
    mediump vec3	square	= (n * n);
    
    // We now roll most of the constants above into the coefficient uniforms, also now compressed into fewer vectors...
    mediump vec3	result  = lightSHCoeff[0].xyz;
    result += lightSHCoeff[1].xyz * n.x;
    result += lightSHCoeff[2].xyz * n.y;
    result += lightSHCoeff[3].xyz * n.z;
    result += lightSHCoeff[4].xyz * flip.x;
    result += lightSHCoeff[5].xyz * flip.y;
    result += lightSHCoeff[6].xyz * flip.z;
    result += vec3(lightSHCoeff[0].w, lightSHCoeff[1].w, lightSHCoeff[2].w) * (square.z - lightSHCoeff[6].w);
    result += vec3(lightSHCoeff[3].w, lightSHCoeff[4].w, lightSHCoeff[5].w) * (square.x - square.y);
    
    result = max(result, 0.0);
    
    return result;
}
#endif

//------------------------------------------------------------------------------

#if defined (SHADER_FOG)
mediump float calculateFogValue(mediump vec4 posModel, mediump vec4 fogPlaneY, mediump vec4 fogPlaneZ, mediump float fogDensity)
{
    mediump float	fogValue	= max(dot(fogPlaneY, posModel), dot(fogPlaneZ, posModel));
    fogValue	= max(fogValue, 0.0);
    
#if defined (SHADER_FOG_EXP2)
#define LOG2 1.442695
    mediump float	result		= exp2(-(fogValue*fogValue*LOG2));
    result		= clamp(result, 0.0, 1.0);
    result		= (1.0 - result) * fogDensity;
#else
    mediump float	result		= min(fogValue, 1.0) * fogDensity;
#endif
    
    return result;
}
#endif

//------------------------------------------------------------------------------

#if defined (SHADER_FOG)
mediump vec4 calculateFogColour(mediump vec3 fogColour, mediump float fogValue)
{
    mediump vec4	result = vec4((uniform_fogColour.xyz * fogValue), (1.0 - fogValue));
    
    return result;
}
#endif

//------------------------------------------------------------------------------

#if defined (SHADER_FOG) && defined (SHADER_FOG_SUN_HAZE) && defined (SHADER_LIGHT)
mediump vec4 calculateFogColourWithSunHaze(mediump vec3 fogColour, mediump float fogValue, mediump vec4 sunHazeFogColour, mediump vec3 eye, mediump vec3 light)
{
    mediump float	hazeFactor	= max(0.0, dot(light, eye));
    hazeFactor	= (hazeFactor * hazeFactor) * sunHazeFogColour.w;//pow(hazeFactor, 8.0);
    mediump vec3	hazeColour	= (sunHazeFogColour.xyz * hazeFactor) + (fogColour.xyz * (1.0 - hazeFactor));
    
    mediump vec4	result		= vec4((hazeColour.xyz * fogValue), (1.0 - fogValue));
    
    return result;
}
#endif

//------------------------------------------------------------------------------

void main()
{
	float	sinAngle = sin(rotScale.z);
	float	cosAngle = cos(rotScale.z);
	
	vec2 offsetPos;
	offsetPos.x  = rotScale.x * cosAngle + rotScale.y * sinAngle; 
	offsetPos.y  = rotScale.y * cosAngle - rotScale.x * sinAngle;
	
    vec4 posWorld = vec4(position.xyz, 1);
    vec4 offsetWorld = viewWorld * vec4(offsetPos, 0, 0);
	posWorld += offsetWorld;
	
	gl_Position = viewScreen * posWorld;
	
	texCoordHueVarying = texCoord;

    vec4 rgba = colour * (1.0/255.0);
    
    colorVarying = rgba;
    
    //--------------------------------------------------------------------------
    // Lighting...
    //--------------------------------------------------------------------------
    
#if defined (SHADER_LIGHT)
    
    mediump vec3	lighting = (ambientLightColour.xyz * (1.0/255.0));//vec3(0.0, 0.0, 0.0);
    
    // Experimental lighting...
    vec3 normalWorldCorner = normalize(offsetWorld.xyz) * lightingSettings.y;
    vec3 normalWorldFace = viewWorld[2].xyz * lightingSettings.z;
    vec3 normalWorldUp = vec3(0.0, 1.0, 0.0) * lightingSettings.w;
    
//    normalWorld = normalize(normalWorld + vec3(0.0, 0.0, 1.0));
//    vec3 normalWorld = normalize(normalWorldCorner + (normalWorldFace * 0.5));
    vec3 normalWorld = normalize(normalWorldCorner + normalWorldFace + normalWorldUp);

#if defined (SHADER_LIGHT_SH)
    lighting += calculateSphericalHarmonics(normalWorld);

    // Directional light...
    mediump vec3	lightWorld	= lightDirection;
    mediump float	normalLightDot	= clamp(dot(lightWorld, normalWorld), 0.0, 1.0);
    lighting += (lightColour * normalLightDot);
#endif
    
#if defined (SHADER_LIGHT_CLASSIC)
//    vec3 lightDirWorld = uniform_lightDirection;//vec3(0.0, 1.0, 0.0);
//    float lightDot = dot(normalWorld, lightDirWorld);
//    colorVarying.xyz *= lightDot;
//    colorVarying.xyz = offsetWorld.xyz;
//    colorVarying.xyz = vec3(lightDot, lightDot, lightDot);

//    colorVarying.xyz = ((lightDot * uniform_lightColour) + uniform_lightAmbient);

    vec4	normalLightDots	= calculateClassicLightingDots(vec4(normalWorld, 0.0), lightDirectionMat);
    
    lighting += (lightColourMat * normalLightDots).xyz;
#endif

#if !defined (SHADER_GAMMA_CORRECT)
    lighting = pow(lighting, vec3(1.0/2.2));
#endif

    lighting = min(lighting, 1.0);
    
    // Pre-multiply alpha...
//    lighting *= colorVarying.w;
    lighting *= colorVarying.xyz;

    colorVarying.xyz = (colorVarying.xyz * (1.0 - lightingSettings.x)) + (lighting * lightingSettings.x);
 
#endif //SHADER_LIGHT

    //--------------------------------------------------------------------------
    // Fog...
    //--------------------------------------------------------------------------
    
#if defined (SHADER_FOG)
    mediump float	fogValue = calculateFogValue(posWorld, uniform_fogPlaneY, uniform_fogPlaneZ, uniform_fogColour.w);
    
#if defined (SHADER_FOG_SUN_HAZE) && defined (SHADER_LIGHT)
#if defined (SHADER_LIGHT_CLASSIC)
    mediump vec3	lightWorld	= vec3(lightDirectionMat[0].x, lightDirectionMat[1].x, lightDirectionMat[2].x);
#endif
    mediump vec3	eyeWorld	= -viewWorld[2].xyz;
    varying_fogColour = calculateFogColourWithSunHaze(uniform_fogColour.xyz, fogValue, uniform_sunHazeColourFog, eyeWorld, lightWorld);
#else
    varying_fogColour = calculateFogColour(uniform_fogColour.xyz, fogValue);
#endif
    
    varying_fogColour.xyz *= colorVarying.w;
//    varying_fogColour.xyz = (vec3(1.0, 1.0, 1.0) * colorVarying.w);
//    varying_fogColour.w = 0.0;
#endif
}

//------------------------------------------------------------------------------
