//
//  Shader_Blitter_Textured.fsh
//
//  Copyright (c) 2012 Midoki Ltd. All rights reserved.
//

#define SHADER_FOG

varying lowp vec4 colorVarying;
varying mediump vec4 texCoordHueVarying;

#if defined (SHADER_FOG)
varying mediump	vec4 varying_fogColour;
#endif

uniform sampler2D baseTexture;

mediump vec3 remapHue(mediump vec3 rgb, mediump float h)
{
	mediump float cMin = min(min(rgb.x, rgb.y), rgb.z);
	mediump float cMax = max(max(rgb.x, rgb.y), rgb.z);
	mediump float v = (cMax + 1.0e-4);
	mediump float s = ((cMax - cMin) / v);
	
	mediump vec4 K = vec4(1.0, 2.0/3.0, 1.0/3.0, 3.0);
	mediump vec3 p2 = abs(fract(h + K.xyz) * 6.0 - K.www);
	
	return (v * mix(K.xxx, clamp(p2 - K.xxx, 0.0, 1.0), s));
}

void main()
{
	mediump vec4 texel = texture2D(baseTexture, texCoordHueVarying.xy);
	
	mediump vec3 texelHue = remapHue(texel.xyz, texCoordHueVarying.z);

	texel.xyz = (texel.xyz * (1.0 - texCoordHueVarying.w)) + (texelHue.xyz * texCoordHueVarying.w);
	
    mediump vec4 finalRGBA = (texel * colorVarying);
    
#if defined (SHADER_FOG)
    finalRGBA.xyz = (finalRGBA.xyz * varying_fogColour.w) + (varying_fogColour.xyz * texel.w);
#endif

    gl_FragColor = finalRGBA;
}
