//
//  Shader_Blitter_3d_Untextured.vsh
//
//  Copyright (c) 2012 Midoki Ltd. All rights reserved.
//

//------------------------------------------------------------------------------

#define SHADER_DISSOLVE

#define lerp(x,y,a) mix(x,y,a)

#define DECOMPRESSION_MULTIPLIER_POSITION   (1.0 / 32767.0)
#define DECOMPRESSION_MULTIPLIER_UV         (1.0 / 1024.0)

//------------------------------------------------------------------------------

attribute highp vec4 attrib_position;
attribute highp vec4 attrib_rgba;
attribute highp vec2 attrib_uv0;

//------------------------------------------------------------------------------

uniform highp mat4 uniform_modelViewProjectionMatrix;

#if defined (SHADER_DISSOLVE)
uniform mediump float uniform_animationValue;
uniform mediump vec4 uniform_emissiveColour;
#endif

//------------------------------------------------------------------------------

varying lowp vec4 varying_colour;
varying mediump vec2 varying_uv0;

#if defined (SHADER_DISSOLVE)
varying mediump vec2 varying_dissolve;
#endif

//------------------------------------------------------------------------------

void main()
{
    varying_colour = vec4(attrib_rgba.xyz, 1.0);

#if defined (SHADER_DISSOLVE)
    // Adjust the animation value to so that the actual 0-1 range is clear of the line width...
    mediump float lineWidth = uniform_emissiveColour.w;
    mediump float lineMax = (uniform_animationValue * (1.0 + lineWidth));
    mediump float lineMin = (lineMax - lineWidth);

    varying_dissolve = vec2(lineMin, lineMax);
#endif

    varying_uv0 = (attrib_uv0 * DECOMPRESSION_MULTIPLIER_UV);
    
    highp vec3 posScaled   = (attrib_position.xyz * (attrib_position.w * DECOMPRESSION_MULTIPLIER_POSITION));
    highp vec4 posScaled4  = vec4(posScaled, 1.0);

    gl_Position = (uniform_modelViewProjectionMatrix * posScaled4);
}

//------------------------------------------------------------------------------
