//
//  Shader_Blitter_Untextured.fsh
//
//  Copyright (c) 2012 Midoki Ltd. All rights reserved.
//

varying lowp vec4 varying_colour;
varying highp vec3 varying_normal;
varying highp vec3 varying_sunVector;

uniform highp vec3 uniform_sunColour;
uniform highp vec4 uniform_sunHazeColour;
uniform highp vec3 uniform_sunSettings;
uniform highp vec4 uniform_horizonColourBlend;
uniform highp vec4 uniform_horizonLightning;

void main()
{
	highp float sunInner = uniform_sunSettings.x;
	highp float sunOuter = uniform_sunSettings.y;
	highp float hazeOuter = uniform_sunSettings.z;
	highp float hazeIntensity = uniform_sunHazeColour.w;

	highp vec3 normalN = normalize(varying_normal);
	
	highp float sunDot = max(0.0, dot(normalN, varying_sunVector));

	mediump vec3 finalRGB = varying_colour.xyz;

	mediump float hazeFactor = max(0.0, ((sunDot - hazeOuter) / (1.0 - hazeOuter))) * hazeIntensity;
	finalRGB = (finalRGB * (1.0 - hazeFactor)) + (uniform_sunHazeColour.xyz * hazeFactor);

	mediump float sunFactor = clamp(((sunDot - sunOuter) / (sunInner - sunOuter)), 0.0, 1.0);
	finalRGB = (finalRGB * (1.0 - sunFactor)) + (uniform_sunColour.xyz * sunFactor);
	
	highp float lightningDot = max(0.0, -dot(normalN, uniform_horizonLightning.xyz));
	mediump float lightningFactor = clamp(((lightningDot - (1.0 - uniform_horizonLightning.w)) / uniform_horizonLightning.w), 0.0, 1.0);
	finalRGB += ((lightningFactor * lightningFactor) * uniform_horizonLightning.w * 0.5);
	
    mediump float grey = dot(vec3(0.3, 0.59, 0.51), vec3(finalRGB));
    mediump vec3 tintedRGB = vec3(grey, grey, grey) * uniform_horizonColourBlend.xyz;
    finalRGB *= (1.0 - uniform_horizonColourBlend.w);
    finalRGB += (tintedRGB * uniform_horizonColourBlend.w);
    
    gl_FragColor = vec4(finalRGB, 1.0);
}
