//
//  Shader_ES300_Base.fsh
//  CityBuilder
//
//  Created by Lee Metcalfe on 23/06/2014.
//  Copyright (c) 2011 Midoki Ltd. All rights reserved.
//
//------------------------------------------------------------------------------

//==============================================================================
// GL specific defines...
//==============================================================================

//#define SHADER_TYPE_VERTEX
#define SHADER_TYPE_FRAGMENT

// Defines and macros to allow the shader code to be compiled for ES2 and ES3...
#if defined ES_300

#define ATTRIBUTE				in
#if defined (SHADER_TYPE_VERTEX)
#define VARYING                 out
#else
#define VARYING                 in
#endif
#define UNIFORM					uniform
#define FRAGCOLOUR				out_fragColour
#define SAMPLER2D				uniform sampler2D
#define SAMPLER2DSHADOW			uniform mediump sampler2DShadow
#define TEXTURE2D(a,b)			texture(a,b)
#define TEXTURE2DSHADOW(a,b)	texture(a,b.xyz/b.w)

#if defined (ANDROID)
#define TEXTURE2DBIAS(a,b,c)	texture(a,b)
#else
#define TEXTURE2DBIAS(a,b,c)	texture(a,b,c)
#endif

#else

#define ATTRIBUTE				attribute
#define VARYING                 varying
#define UNIFORM					uniform
#define FRAGCOLOUR				gl_FragColor
#define SAMPLER2D				uniform sampler2D
#define TEXTURE2D(a,b)			texture2D(a,b)

#if defined (OSX)
#define SAMPLER2DSHADOW			uniform sampler2DShadow
#define TEXTURE2DSHADOW(a,b)	shadow2DProj(a,b).x
#else
#define SAMPLER2DSHADOW			uniform sampler2D
#define TEXTURE2DSHADOW(a,b)	texture2DProj(a,b).x
#endif

#if defined (ANDROID)
#define TEXTURE2DBIAS(a,b,c)	texture2D(a,b)
#else
#define TEXTURE2DBIAS(a,b,c)	texture2D(a,b,c)
#endif

#endif

//------------------------------------------------------------------------------

//==============================================================================
// Common defines...
//==============================================================================

#define PI		3.1415926
#define PI20	(PI*2.0)
#define PI05	(PI*0.5)

//------------------------------------------------------------------------------
// Some special cases...
//------------------------------------------------------------------------------

#if defined (SHADER_MARSH_WATER)
#undef SHADER_TEXTURE_NORMAL
#undef SHADER_TEXTURE_NORMAL2
#undef SHADER_TEXTURE_ANIMUV
#undef SHADER_TEXTURE_REFLECTION_PLANE

#define SHADER_TEXTURE_NORMAL
#define SHADER_TEXTURE_NORMAL2
#define SHADER_TEXTURE_ANIMUV
#define SHADER_TEXTURE_REFLECTION_PLANE
#endif

//------------------------------------------------------------------------------

// If lighting is enabled, also force the fixed colour to be used...
#if defined (SHADER_LIGHT)
#define SHADER_FIXED_COLOUR
#endif

// Simplified define for the shadowmap case...
#if defined (SHADER_TEXTURE) && defined (SHADER_TEXTURE_MULTI) && defined (SHADER_TEXTURE_MULTI_SHADOWMAP)
#define SHADER_TEXTURE_SHADOWMAP
#endif

#define SHADER_LIGHT_POINT

//------------------------------------------------------------------------------

//==============================================================================
// Fragment shader defines...
//==============================================================================

#if defined (SHADER_REFLECTIVE_SURFACE)
#define REFLECTION_SCREEN_X_SCALE   (1.0)
#define REFLECTION_SCREEN_Y_SCALE   (1.0)
#else
#define REFLECTION_SCREEN_X_SCALE   (0.9)
#define REFLECTION_SCREEN_Y_SCALE   (0.8)
#endif

//------------------------------------------------------------------------------

#if !defined (SHADER_SHADOW_DITHER_DISABLE)
#define SHADER_SHADOW_DITHER
#endif

//------------------------------------------------------------------------------

//==============================================================================
// Uniforms...
//==============================================================================

#if defined (SHADER_ADD_COLOUR)
UNIFORM mediump		vec4 uniform_addColour;
#endif

#if defined (SHADER_TEXTURE_REFLECTION_SPHERE) || defined (SHADER_TEXTURE_REFLECTION_LATLONG)|| defined (SHADER_TEXTURE_REFLECTION_PLANE)
UNIFORM mediump		vec4 uniform_reflectionColour;
#endif

#if defined (SHADER_TEXTURE_EMISSIVE) || defined (SHADER_DISSOLVE)
UNIFORM mediump		vec4 uniform_emissiveColour;
#endif

#if defined (SHADER_HSV_ADJUST)
UNIFORM mediump		vec3 uniform_hsvAdjust;
#elif defined (SHADER_REMAP_HUE)
UNIFORM mediump		float uniform_remapHue;
#endif

#if defined (SHADER_MARSH_WATER)
UNIFORM highp		mat4 uniform_viewProjectionMatrix;
UNIFORM highp		vec3 uniform_lightDirection;
UNIFORM highp		vec3 uniform_eyePosition;
UNIFORM highp		vec4 uniform_oceanWaterSettings;
#endif

#if defined (SHADER_REFLECTIVE_SURFACE)
UNIFORM highp        vec3 uniform_eyePosition;
#endif

#if defined (SHADER_SNOW)
UNIFORM highp		vec3 uniform_lightDirection;
UNIFORM highp		vec3 uniform_lightColour;
UNIFORM highp		vec3 uniform_eyePosition;
UNIFORM highp		vec4 uniform_specularRGBPow;
//UNIFORM mediump   vec3 uniform_horizonSunLightDirection;
UNIFORM mediump     vec3 uniform_specularDirection;

UNIFORM mediump		vec3 uniform_snowSettings;
#endif

#if defined (SHADER_LIGHT) && defined (SHADER_LIGHT_SPECULAR)
UNIFORM mediump     vec4 uniform_specularRGBPow;
#if defined (SHADER_PER_PIXEL_SPECULAR)
UNIFORM highp		vec3 uniform_eyePosition;
#endif
#endif

#if defined (SHADER_SNOW_FRESNEL) || defined (SHADER_FRESNEL_TRANS)
UNIFORM mediump     vec4 uniform_fresnelColour1;
UNIFORM mediump     vec4 uniform_fresnelColour2;
#endif

#if defined (SHADER_DEPTH_ONLY)
#if defined (SHADER_DEPTH_ONLY_ALPHA_FADE)
UNIFORM mediump     float uniform_shadowAlpha;
UNIFORM mediump     float uniform_bayer4x4[16];
//UNIFORM mediump     float uniform_bayer8x8[64];
#endif
#endif

#if defined (SHADER_FRESNEL_TRANS)
UNIFORM mediump     vec4 uniform_fresnelTransSettings;
UNIFORM highp       vec3 uniform_eyePosition;
#endif

//------------------------------------------------------------------------------

#if defined (SHADER_TEXTURE)
SAMPLER2D uniform_texture0;
#if defined (SHADER_TEXTURE_MULTI) || defined (SHADER_CELSHADE_2TEX)
SAMPLER2D uniform_texture1;
#endif
#endif

#if defined (SHADER_TEXTURE_EMISSIVE)
SAMPLER2D uniform_emissivemap;
#endif

#if defined (SHADER_TEXTURE_REFLECTION_SPHERE) || defined (SHADER_TEXTURE_REFLECTION_LATLONG)
SAMPLER2D uniform_reflectionmap;
#endif

#if defined (SHADER_TEXTURE_REFRACTION) || defined (SHADER_TEXTURE_REFLECTION_PLANE) || defined (SHADER_MAP_LOCKED_MASK)
SAMPLER2D uniform_refractionmap;
#endif

#if defined (SHADER_TEXTURE_NORMAL)
SAMPLER2D uniform_normalmap;
#endif
#if defined (SHADER_TEXTURE_NORMAL2)
SAMPLER2D uniform_normalmap2;
//SAMPLER2D	uniform_alphamap;
#endif

#if defined (SHADER_DISSOLVE)
SAMPLER2D    uniform_alphamap;
#endif

#if defined (SHADER_SHADOW_RECEIVE)
SAMPLER2DSHADOW uniform_shadowmap;
#endif

#if defined (SHADER_LIGHT) && defined (SHADER_LIGHT_SPECULAR)
#if defined (SHADER_TEXTURE_SPECULAR)
SAMPLER2D uniform_specularmap;
#endif
#endif

#if defined (SHADER_MAP_CLOUD_SHADOWS)
SAMPLER2D uniform_alphamap;
#endif

#if defined (SHADER_MARSH_WATER) || defined (SHADER_MAP_CLOUD_SHADOWS)
SAMPLER2D uniform_ambientmap;
#endif

#if defined (SHADER_SNOW_NORMALMAP)
SAMPLER2D uniform_normalmap;
#endif

#if defined (SHADER_SNOW)
SAMPLER2D uniform_specularmap;
#endif

//------------------------------------------------------------------------------

//==============================================================================
// Vertex shader outputs...
//==============================================================================

#if defined (SHADER_LIGHT) || defined (SHADER_VCOL) || defined (SHADER_FIXED_COLOUR) || defined (SHADER_MARSH_WATER) || defined (SHADER_TEXTURE_MULTI_VRED)
	VARYING lowp    vec4 varying_rgba;
#endif

//------------------------------------------------------------------------------
// UV's - the conditions here match the code in the vertex shader, this is required so we can vary precision for the animated uv's...

#if defined (SHADER_TEXTURE) || defined (SHADER_TEXTURE_EMISSIVE) || defined (SHADER_SNOW_NORMALMAP)
#if defined (SHADER_WATER_FOAM)

VARYING lowp	vec2 varying_uv0;
VARYING mediump	vec4 varying_uv1uv2;

#elif defined (SHADER_MARSH_WATER)

VARYING mediump	vec2 varying_uv0;
VARYING mediump	vec4 varying_uv1uv2;

#elif defined (SHADER_TEXTURE_ANIMUV)

// Animated (scrolling) uv's...
VARYING mediump	vec2 varying_uv0;
#if defined (SHADER_TEXTURE_MULTI_UV2)
VARYING mediump	vec2 varying_uv1;
#endif
#if defined (SHADER_SPECULAR_ALPHA)
VARYING lowp	vec2 varying_uv1;
#endif

#elif defined (SHADER_TEXTURE_ANIMUVFRAME) || defined (SHADER_TEXTURE_ANIMUVFRAMEBLEND)

// Animated (frame cycling) uv's...
VARYING mediump vec2 varying_uv0;
#if defined (SHADER_TEXTURE_ANIMUVFRAMEBLEND)
VARYING highp   vec3 varying_animUVFrameBlend;
#endif

#elif defined (SHADER_TEXTURE_EMISSIVE) || defined (SHADER_SNOW_NORMALMAP)

VARYING lowp    vec2 varying_uv0;

#else

// Static uv's...
#if defined (SHADER_TEXTURE)
VARYING lowp	vec2 varying_uv0;
#if defined (SHADER_TEXTURE_MULTI_UV2)
VARYING lowp	vec2 varying_uv1;
#endif
#endif

#endif  //defined (SHADER_WATER_FOAM)
#endif  //defined (SHADER_TEXTURE) || defined (SHADER_TEXTURE_EMISSIVE)

//------------------------------------------------------------------------------

#if defined (SHADER_TEXTURE_REFLECTION_SPHERE) || defined (SHADER_TEXTURE_REFLECTION_LATLONG)
VARYING mediump vec4 varying_reflectionUV;
#endif

#if defined (SHADER_TEXTURE_REFRACTION)
VARYING mediump vec2 varying_refractionUV;
#endif

#if defined (SHADER_MAP_CLOUD_SHADOWS)
VARYING lowp	vec4 varying_cloud_uv0uv1;
#endif

//------------------------------------------------------------------------------

#if defined (SHADER_LIGHT) && defined (SHADER_LIGHT_SPECULAR)
////VARYING mediump	vec3 varying_specLightVec;
////VARYING mediump	vec3 varying_specNormalVec;
////VARYING mediump	vec3 varying_specEyeVec;
//VARYING mediump	vec3 varying_specRefVec;
//VARYING mediump	vec3 varying_specEyeVec;
#if defined (SHADER_PER_PIXEL_SPECULAR)
VARYING mediump	vec4 varying_specLightVec;
VARYING highp	vec3 varying_specPosWorld;
#else
VARYING mediump	vec4 varying_specRefVec;
#endif
#if !defined (SHADER_SPECULAR_NORMAL_UP)
VARYING mediump	vec3 varying_specNormal;
#endif
#endif

#if defined (SHADER_SHADOW_RECEIVE) || defined (SHADER_MAP_CLOUD_SHADOWS) || defined (SHADER_SNOW) || defined (SHADER_TEXTURE_SHADOWMAP)
VARYING lowp    vec3 varying_shadowAmbient;
#endif
#if defined (SHADER_SHADOW_RECEIVE)
VARYING highp	vec4 varying_shadowUVZW;
#endif

#if defined (SHADER_FOG)
VARYING mediump	vec4 varying_fogColour;
#endif

#if defined (SHADER_TEXTURE_REFLECTION_PLANE) //|| (defined (SHADER_SHADOW_RECEIVE) && defined (SHADER_SHADOW_DITHER))
VARYING highp vec4 varying_posScreen;
#endif

#if defined (SHADER_MARSH_WATER) || defined (SHADER_REFLECTIVE_SURFACE) || defined (SHADER_FRESNEL_TRANS)
VARYING highp vec4 varying_posWorld;
#endif

#if defined (SHADER_SNOW)
VARYING mediump	vec3 varying_snowNormal;
VARYING highp	vec4 varying_snowPosWorld_Deform;
#endif

#if defined (SHADER_MAP_LOCKED_MASK)
VARYING highp   vec4 varying_mapLockedMaskPosScreen;
#endif

#if defined (SHADER_DISSOLVE)
VARYING mediump vec2 varying_dissolve;
#endif

#if defined (SHADER_FRESNEL_TRANS)
VARYING mediump vec3 varying_fresnelTransNormal;
#endif

//------------------------------------------------------------------------------

//==============================================================================
// Fragment shader output (GL only)...
//==============================================================================

#if defined (OSX) || defined (ANDROID)
#if defined (SHADER_SHADOW_RECEIVE) && defined (SHADER_SHADOW_DITHER)
#if !defined ES_300
VARYING mediump vec4 gl_FragCoord;
#endif
#endif
#endif

#if defined ES_300
out lowp vec4 out_fragColour;
#endif

//------------------------------------------------------------------------------

#if defined (SHADER_HSV_ADJUST)
mediump vec3 rgb2hsv(vec3 c)
{
	mediump vec4 K = vec4(0.0, -1.0 / 3.0, 2.0 / 3.0, -1.0);
	mediump vec4 p = mix(vec4(c.bg, K.wz), vec4(c.gb, K.xy), step(c.b, c.g));
	mediump vec4 q = mix(vec4(p.xyw, c.r), vec4(c.r, p.yzx), step(p.x, c.r));
	
	mediump float d = q.x - min(q.w, q.y);
	mediump float e = 1.0e-4;
	return vec3(abs(q.z + (q.w - q.y) / (6.0 * d + e)), d / (q.x + e), q.x);
}
#endif

//------------------------------------------------------------------------------

#if defined (SHADER_HSV_ADJUST)
mediump vec3 hsv2rgb(vec3 c)
{
	mediump vec4 K = vec4(1.0, 2.0 / 3.0, 1.0 / 3.0, 3.0);
	mediump vec3 p = abs(fract(c.xxx + K.xyz) * 6.0 - K.www);
	return c.z * mix(K.xxx, clamp(p - K.xxx, 0.0, 1.0), c.y);
}
#endif

//------------------------------------------------------------------------------

#if defined (SHADER_REMAP_HUE)
mediump vec3 remapHue(mediump vec3 rgb, mediump float h)
{
	mediump float cMin = min(min(rgb.x, rgb.y), rgb.z);
	mediump float cMax = max(max(rgb.x, rgb.y), rgb.z);
	mediump float v = (cMax + 1.0e-4);
	mediump float s = ((cMax - cMin) / v);

	mediump vec4 K = vec4(1.0, 2.0 / 3.0, 1.0 / 3.0, 3.0);
	mediump vec3 p2 = abs(fract(h + K.xyz) * 6.0 - K.www);
	
	return (v * mix(K.xxx, clamp(p2 - K.xxx, 0.0, 1.0), s));
}
#endif

//------------------------------------------------------------------------------

#if defined (SHADER_SNOW_FRESNEL) || defined (SHADER_REFLECTIVE_SURFACE)
mediump float fresnel(mediump vec3 CameraVector, mediump vec3 Normal, mediump float Min, mediump float Max, mediump float Strength)
{
	mediump float OneMinusOp = (1.0 - dot(CameraVector, Normal));
	mediump float SubOp = clamp(((OneMinusOp) - Min), 0.0, 1.0);
	mediump float SubOp1511 = (max(Max, Min) - Min);
	mediump float SatOp = clamp((SubOp / SubOp1511), 0.0, 1.0);
	mediump float MulOp = (SatOp * Strength);
	
	return MulOp;
}
#endif

//------------------------------------------------------------------------------

void main()
{
#if defined (SHADER_DEPTH_ONLY)

#if defined (SHADER_DEPTH_ONLY_ALPHA_FADE)
//    // 2x2
//    mediump vec2 index2x2 = sign(max((fract(gl_FragCoord.xy * 0.5) - 0.25), 0.0));
//    mediump float limit = (0.25 + (index2x2.x * 0.50) + (index2x2.y * 0.75) - (index2x2.x * index2x2.y));
//    if ((uniform_shadowAlpha - limit) < 0.0)
//    {
//        discard;
//    }

    // 4x4
    mediump vec2 index4x4 = (max((fract(gl_FragCoord.xy * 0.25) - 0.125), 0.0) * 4.0);
    int index = int(index4x4.x + (index4x4.y * 4.0));
    if ((uniform_shadowAlpha - uniform_bayer4x4[index]) < 0.0)
    {
        discard;
    }

//    // 8x8
//    mediump vec2 index8x8 = (max((fract(gl_FragCoord.xy * 0.125) - 0.0625), 0.0) * 8.0);
//    int index = int(index8x8.x + (index8x8.y * 8.0));
//    if ((uniform_shadowAlpha - uniform_bayer8x8[index]) < 0.0)
//    {
//        discard;
//    }
#endif
    
    FRAGCOLOUR = vec4(1.0, 1.0, 1.0, 1.0);

#else
	
#if defined (SHADER_TEXTURE)
	
    #if defined (SHADER_TEXTURE_OCEAN_OPACITY)
    lowp vec4   albedo  = TEXTURE2DBIAS(uniform_texture0, varying_uv0,  -1.0);
	#elif defined (SHADER_TEXTURE_REFLECTION_PLANE) && !defined (SHADER_REFLECTIVE_SURFACE)
	lowp vec4	albedo	= vec4(1.0, 1.0, 1.0, 1.0);
	#elif defined (SHADER_TEXTURE_ALPHA_ONLY)
	lowp float	alpha	= TEXTURE2DBIAS(uniform_texture0, varying_uv0,  -1.0).w;
	lowp vec4	albedo	= vec4(1.0, 1.0, 1.0, alpha);
	#else
	lowp vec4	albedo	= TEXTURE2DBIAS(uniform_texture0, varying_uv0,  -1.0);
	#endif
	
    #if defined (SHADER_CALCIFY)
    // Custom 'calcify' effect for characters turning to stone, it's a simple greyscale conversion with a few tweaks to brighten and reduce contrast...
    const lowp vec3 calcifyDot = vec3(0.2989, 0.5870, 0.1140);
    lowp float calcifyGreyScale = dot(albedo.xyz, calcifyDot);
    calcifyGreyScale = ((calcifyGreyScale * 0.7) + 0.3);
    calcifyGreyScale = max(calcifyGreyScale, 0.6);
    albedo.xyz = vec3(calcifyGreyScale, calcifyGreyScale, calcifyGreyScale);
    #endif
    
//	#if defined (SHADER_CELSHADE_2TEX)
//	lowp vec4	albedo2	= TEXTURE2DBIAS(uniform_texture1, varying_uv0,  -1.0);
//	#endif
	
	#if defined (SHADER_TEXTURE_PUNCH_THROUGH)
	if (albedo.w < 0.8)
	{
		discard;
	}
	#endif
    
	#if defined (SHADER_WATER_FOAM)
	
		lowp vec4	albedo1	= TEXTURE2DBIAS(uniform_texture0, varying_uv1uv2.xy,  -1.0);
		lowp vec4	albedo2	= TEXTURE2DBIAS(uniform_texture0, varying_uv1uv2.zw,  -1.0);
		lowp float	waterEdge = max(albedo.x, (albedo1.x * albedo2.x));
		albedo = vec4(1.0, 1.0, 1.0, waterEdge);
	
	#elif defined (SHADER_TEXTURE_MULTI)
	
		#if defined (SHADER_TEXTURE_MULTI_UV2)
		lowp vec4 texel2 = TEXTURE2DBIAS(uniform_texture1, varying_uv1, -1.0);
		#else
		lowp vec4 texel2 = TEXTURE2DBIAS(uniform_texture1, varying_uv0, -1.0);
		#endif
		
		#if defined (SHADER_TEXTURE_MULTI_MULTIPLY)
		albedo.xyz = (albedo.xyz * texel2.xyz);
		#elif defined (SHADER_TEXTURE_MULTI_SHADOWMAP)
		// Applied below...
		#elif defined (SHADER_TEXTURE_MULTI_VRED)
        albedo.xyz = (albedo.xyz * (1.0 - varying_rgba.w)) + (texel2.xyz * varying_rgba.w);
        #else
		albedo.xyz = (albedo.xyz * (1.0 - texel2.w)) + (texel2.xyz * texel2.w);
		#endif
	
	#endif
	
#else

	lowp vec4 albedo = vec4(1.0, 1.0, 1.0, 1.0);
	
#endif

#if defined (SHADER_TEXTURE_EMISSIVE)
	lowp vec4 emissive = TEXTURE2D(uniform_emissivemap, varying_uv0) * uniform_emissiveColour;
	albedo.xyz = ((albedo.xyz * (1.0 - emissive.w)) + (emissive.xyz * emissive.w));
#endif
	
#if defined (SHADER_TEXTURE_ANIMUVFRAMEBLEND)
	lowp vec4 texel2 = TEXTURE2DBIAS(uniform_texture0, varying_animUVFrameBlend.xy, -1.0);
//	albedo.xyz = (albedo.xyz * (1.0 - varying_animUVFrameBlend.z)) + (texel2.xyz * varying_animUVFrameBlend.z);
    albedo.xyzw = (albedo.xyzw * (1.0 - varying_animUVFrameBlend.z)) + (texel2.xyzw * varying_animUVFrameBlend.z);
#endif
	
#if defined (SHADER_SNOW)
    mediump vec3	eyeToPos = (varying_snowPosWorld_Deform.xyz - uniform_eyePosition.xyz);
    mediump vec3	eyeDirWorld = normalize(eyeToPos);

#if defined (SHADER_SNOW_NORMALMAP)
	//TODO - move this part to the vertex shader...
    mediump vec3	snowN = normalize(varying_snowNormal.xyz);
	mediump vec3	snowT = normalize(cross(snowN, vec3(0.0, 0.0, 1.0)));
	mediump vec3	snowB = normalize(cross(snowN, snowT));

    mediump float	snowNormalIntensity = uniform_snowSettings.z;

	mediump vec3	normalTexel = (TEXTURE2D(uniform_normalmap, varying_uv0.xy).xyz * 2.0) - 1.0;
	mediump vec3	normalWorld0 = normalize((snowT * normalTexel.x) + (snowB * normalTexel.y) + (snowN * normalTexel.z));
    normalWorld0 = (normalWorld0 * snowNormalIntensity) + (snowN * (1.0 - snowNormalIntensity));
#else
    mediump vec3	normalWorld0 = normalize(varying_snowNormal.xyz);
#endif

    mediump vec4	lightingNormal = vec4(normalWorld0, 0.0);
#endif
    
#if defined (SHADER_LIGHT)  && defined (SHADER_VCOL) && defined (SHADER_REFLECTIVE_SURFACE)
    mediump float reflectionLevel = (varying_rgba.w * albedo.w);
    lowp vec4 diffuse = vec4(varying_rgba.xyz, 1.0);
#elif defined (SHADER_LIGHT) || defined (SHADER_VCOL) || defined (SHADER_FIXED_COLOUR)
	lowp vec4 diffuse = varying_rgba;
#else
	lowp vec4 diffuse = vec4(1.0, 1.0, 1.0, 1.0);
#endif
    
#if defined (SHADER_SNOW)
	mediump float	normalLightDot	= clamp(dot(uniform_lightDirection.xyz, lightingNormal.xyz), 0.0, 1.0);
#if defined (SHADER_SNOW_DEFORMATION)
	normalLightDot *= varying_snowPosWorld_Deform.w;
#endif
	diffuse.xyz = (diffuse.xyz * normalLightDot) + (varying_shadowAmbient.xyz * (1.0 - normalLightDot));
#endif
	
#if defined (SHADER_MAP_CLOUD_SHADOWS)
	mediump vec4 cloudShadowAlpha = TEXTURE2D(uniform_alphamap, varying_cloud_uv0uv1.xy);
	mediump vec4 cloudShadowAmbient = TEXTURE2D(uniform_ambientmap, varying_cloud_uv0uv1.zw);
	mediump float cloudShadow = (cloudShadowAmbient.z * cloudShadowAlpha.z);
#endif

#if defined (SHADER_SHADOW_RECEIVE)

#if defined (SHADER_SHADOW_DITHER) || defined (SHADER_SHADOW_RECEIVE_SOFT)
	mediump float texelSize = ((1.0 / 1024.0) * varying_shadowUVZW.w);
#if defined (SHADER_SHADOW_DITHER)
	// Neat little dithering approach for soft shadows from gpu gems...
    mediump vec2 ditherOffset = sign(fract(gl_FragCoord.xy * 0.5) - 0.25);  // mod
    ditherOffset.y += ditherOffset.x;  // y ^= x in floating point
#endif
#endif
	
#if defined (SHADER_SHADOW_RECEIVE_SOFT)
//	mediump float texelSize = (3.456 / 2048) * varying_shadowUVZW.w;
//	mediump float texelSize = ((((fract((varying_shadowUVZW.x + varying_shadowUVZW.y) * 200000.0) + 0.2) * 8.0) / 2048.0) * varying_shadowUVZW.w);
//	shadow += TEXTURE2DSHADOW(uniform_shadowmap, varying_shadowUVZW + vec4(-texelSize, -texelSize, 0.0, 0.0));
//	shadow += TEXTURE2DSHADOW(uniform_shadowmap, varying_shadowUVZW + vec4(       0.0, -texelSize, 0.0, 0.0));
//	shadow += TEXTURE2DSHADOW(uniform_shadowmap, varying_shadowUVZW + vec4( texelSize, -texelSize, 0.0, 0.0));
//	shadow += TEXTURE2DSHADOW(uniform_shadowmap, varying_shadowUVZW + vec4(-texelSize,        0.0, 0.0, 0.0));
//	shadow += TEXTURE2DSHADOW(uniform_shadowmap, varying_shadowUVZW + vec4( texelSize,        0.0, 0.0, 0.0));
//	shadow += TEXTURE2DSHADOW(uniform_shadowmap, varying_shadowUVZW + vec4(-texelSize,  texelSize, 0.0, 0.0));
//	shadow += TEXTURE2DSHADOW(uniform_shadowmap, varying_shadowUVZW + vec4(       0.0,  texelSize, 0.0, 0.0));
//	shadow += TEXTURE2DSHADOW(uniform_shadowmap, varying_shadowUVZW + vec4( texelSize,  texelSize, 0.0, 0.0));
//	shadow = (shadow * (1.0 / 9.0));
	
	highp vec4 softShadowUVZW1 = varying_shadowUVZW;
	highp vec4 softShadowUVZW2 = varying_shadowUVZW;
	highp vec4 softShadowUVZW3 = varying_shadowUVZW;
	highp vec4 softShadowUVZW4 = varying_shadowUVZW;
	
#if defined (SHADER_SHADOW_DITHER)
	softShadowUVZW1.xy += ((ditherOffset + vec2(-1.5, 0.5)) * texelSize);
	softShadowUVZW2.xy += ((ditherOffset + vec2( 0.5, 0.5)) * texelSize);
	softShadowUVZW3.xy += ((ditherOffset + vec2(-1.5,-1.5)) * texelSize);
	softShadowUVZW4.xy += ((ditherOffset + vec2( 0.5,-1.5)) * texelSize);
#else
	softShadowUVZW1.xy += (vec2(-1.5, 0.5) * texelSize);
	softShadowUVZW2.xy += (vec2( 0.5, 0.5) * texelSize);
	softShadowUVZW3.xy += (vec2(-1.5,-1.5) * texelSize);
	softShadowUVZW4.xy += (vec2( 0.5,-1.5) * texelSize);
#endif
	
	mediump float shadow = TEXTURE2DSHADOW(uniform_shadowmap, softShadowUVZW1);
	shadow += TEXTURE2DSHADOW(uniform_shadowmap, softShadowUVZW2);
	shadow += TEXTURE2DSHADOW(uniform_shadowmap, softShadowUVZW3);
	shadow += TEXTURE2DSHADOW(uniform_shadowmap, softShadowUVZW4);
	shadow *= 0.25;
#elif defined (SHADER_SHADOW_DITHER)
	mediump vec4 softShadowUVZW1 = varying_shadowUVZW;
	softShadowUVZW1.xy += (ditherOffset * texelSize);
	mediump float shadow = TEXTURE2DSHADOW(uniform_shadowmap, softShadowUVZW1);
#else
	mediump float shadow = TEXTURE2DSHADOW(uniform_shadowmap, varying_shadowUVZW);
#endif

#if defined (SHADER_TEXTURE_SHADOWMAP)
	shadow = min(shadow, texel2.x);
#endif
	
	#if defined (SHADER_MAP_CLOUD_SHADOWS)
	shadow *= (1.0 - cloudShadow);
	#endif

	diffuse.xyz = (diffuse.xyz * shadow) + (varying_shadowAmbient.xyz * (1.0 - shadow));
#else
#if defined (SHADER_MAP_CLOUD_SHADOWS)
//	diffuse.xyz *= (1.0 - cloudShadow);
	mediump float shadow = (1.0 - cloudShadow);
	diffuse.xyz = (diffuse.xyz * shadow) + (varying_shadowAmbient.xyz * (1.0 - shadow));
#endif
#if defined (SHADER_TEXTURE_SHADOWMAP)
	mediump float shadow = texel2.x;
	diffuse.xyz = (diffuse.xyz * shadow) + (varying_shadowAmbient.xyz * (1.0 - shadow));
#endif
#endif

	// Snow fresnel white/blue colour shift...
#if defined (SHADER_SNOW)
#if defined (SHADER_TEXTURE_MULTI_UV2)
	mediump vec4 snowSpecularTexel = TEXTURE2D(uniform_specularmap, varying_uv1);
#else
    mediump vec4 snowSpecularTexel = TEXTURE2D(uniform_specularmap, varying_uv0);
#endif
	
#if defined (SHADER_SNOW_SPARKLE)
    mediump float snowWetness = uniform_snowSettings.x;
    mediump float snowSparkleMask = ((snowSpecularTexel.x * 50.0) + snowWetness);
#else
    mediump float snowSparkleMask = snowSpecularTexel.x;
#endif
//	mediump float snowSparkleMask = ((albedo.x * 50.0) + 0.0825688);
	//albedo.xyz = vec3(1.0, 1.0, 1.0);
	
#if defined (SHADER_SNOW_FRESNEL)
    mediump float fresMin = 0.0;
	mediump float fresMax = 1.0;
	mediump float fresStr = 1.0;//1.5;
	mediump float fresVal = fresnel(-eyeDirWorld, lightingNormal.xyz, fresMin, fresMax, fresStr);
	//	mediump vec3 fresCol0 = vec3(0.4117,0.7298,1.0);
    mediump vec3 fresCol0 = uniform_fresnelColour1.xyz;//vec3(0.6,0.8,1.0);
    mediump vec3 fresCol1 = uniform_fresnelColour2.xyz;//vec3(1.0, 1.0, 1.0);
	mediump vec3 fresColour = ((fresCol0 * (1.0 - fresVal)) + (fresCol1 * fresVal));
	//	mediump vec3 fresColour = ((uniform_fresnelColour1.xyz * (1.0 - fresVal)) + (uniform_fresnelColour2.xyz * fresVal));
	albedo.xyz = fresColour;
#endif
#endif
    
    // Fresnel trans effect...
#if defined (SHADER_FRESNEL_TRANS)
    mediump vec3 eyeToPos = (varying_posWorld.xyz - uniform_eyePosition.xyz);
    mediump vec3 eyeDirWorld = normalize(eyeToPos);
    mediump vec3 fresnelTransNormal = normalize(varying_fresnelTransNormal.xyz);

    //mediump float fresVal0 = fresnel(-eyeDirWorld, fresnelTransNormal.xyz, uniform_fresnelTransSettings.x, uniform_fresnelTransSettings.y, 1.0);
    //mediump float fresVal1 = fresnel(-eyeDirWorld, fresnelTransNormal.xyz, uniform_fresnelTransSettings.x, uniform_fresnelTransSettings.y, 1.0);
   
    // Perform the fresnel operation in place so we can optimise it to produce two results...
    mediump float OneMinusOp = (1.0 - dot(-eyeDirWorld, fresnelTransNormal.xyz));
    mediump float minA = uniform_fresnelTransSettings.x;
    mediump float maxA = uniform_fresnelTransSettings.y;
    mediump float minB = uniform_fresnelTransSettings.z;
    mediump float maxB = uniform_fresnelTransSettings.w;
    mediump float SubOpA = clamp(((OneMinusOp) - minA), 0.0, 1.0);
    mediump float SubOpB = clamp(((OneMinusOp) - minB), 0.0, 1.0);
    mediump float SubOp1511A = (max(maxA, minA) - minA);
    mediump float SubOp1511B = (max(maxB, minB) - minB);
    mediump float fresValA = clamp((SubOpA / SubOp1511A), 0.0, 1.0);
    mediump float fresValB = clamp((SubOpB / SubOp1511B), 0.0, 1.0);

    mediump vec3 fresColA = uniform_fresnelColour1.xyz;
    mediump vec3 fresColB = uniform_fresnelColour2.xyz;
    mediump vec3 fresColour = ((fresColA * fresValA) + (fresColB * fresValB));
    
    albedo.xyz = fresColour;
#endif
    
	// SPECIAL: Colourize the texture by combining the intensity of the pixel with the colour...
#if defined (SHADER_VCOL_COLOURIZE)
	lowp float intensity = max(max(albedo.x, albedo.y), albedo.z) * 1.5;
//	lowp vec4 finalRGBA = vec4((intensity * diffuse.xyz), albedo.w);
	lowp vec4 finalRGBA = vec4(intensity, intensity, intensity, albedo.w) * diffuse;
#else
	lowp vec4 finalRGBA = (albedo * diffuse);
#endif
	
#if defined (SHADER_SPECULAR_ALPHA)
    finalRGBA.w = diffuse.w;
#endif
    
	// SPECIAL: Added colour tint for unit special effects...
#if defined (SHADER_ADD_COLOUR)
	mediump float intensity = dot(finalRGBA.xyz, vec3(0.3, 0.5, 0.2));
	mediump vec3 tintedRGB = (uniform_addColour.xyz + vec3(intensity, intensity, intensity));
	finalRGBA.xyz = (finalRGBA.xyz * (1.0 - uniform_addColour.w)) + (tintedRGB * uniform_addColour.w);
#endif
	
	//--------------------------------------------------------------------------
	// Lighting...
	//--------------------------------------------------------------------------
	//LM: I've moved part of the water reflection code earlier so the normal calculated from the normal maps can be used the in the specular lighting calculation...
#if defined (SHADER_TEXTURE_REFLECTION_PLANE) && defined (SHADER_MARSH_WATER)
	mediump float	mask = TEXTURE2D(uniform_ambientmap, varying_uv0).x;
	mediump float   normalHeight = varying_rgba.w;// * mask;
	mediump float   reflectionLevel = mask;//(1.0 - mask.x);//(1.0 - normalHeight);

	normalHeight    *= uniform_oceanWaterSettings.z;
	reflectionLevel *= uniform_oceanWaterSettings.w;

	mediump vec3	lightDirWorld = vec3(normalize(uniform_lightDirection));
	
	mediump vec3	eyeToPos = vec3(varying_posWorld.xyz - uniform_eyePosition.xyz);
	mediump vec3	eyeDirWorld = normalize(eyeToPos);
	
	mediump float	eyeToPosDist = length(eyeToPos);
	mediump float	distortionScale = (clamp(eyeToPosDist, 4.0, 16.0) / 4.0);
	
	// METHOD 2: Perform a proper reflection calculation, this can be manipulated with a normal map...
	mediump vec3	normalWorld0 = (TEXTURE2D(uniform_normalmap, varying_uv1uv2.xy).xyz * 2.0) - 1.0;
	mediump vec3	normalWorld1 = (TEXTURE2D(uniform_normalmap2, varying_uv1uv2.zw).xyz * 2.0) - 1.0;
//	mediump vec3	normalWorldA = normalize((normalWorld0 * 0.5) + (normalWorld1 * 0.5));
	mediump vec3	normalWorldA = normalize(normalWorld0 + normalWorld1);

	// Correct the 'space' of the normal map...
	mediump vec3	normalWorld = vec3(normalWorldA.x, normalWorldA.z, -normalWorldA.y);

	// Apply the normal height to the normal...
	normalWorld.xz *= normalHeight;
	normalWorld = normalize(normalWorld);
#endif

#if defined (SHADER_LIGHT) && defined (SHADER_LIGHT_SPECULAR)
//	// CORRECT: Full per-pixel specular...
//	mediump	float	specCoeff = pow(max(0.0, dot(varying_specEyeVec, reflect(normalize(varying_specLightVec), normalize(varying_specNormalVec)))), 4.0);
	
//	// FAST: Perform the reflect per vertex and interpolate (and normalize) per pixel, this actually looks identical...
//	mediump	float	specCoeff = pow(max(0.0, dot(normalize(varying_specEyeVec), normalize(varying_specRefVec))), uniform_specularRGBPow.w/4.0);

#if defined (SHADER_MARSH_WATER)
    mediump vec3    specNormal = normalWorld;
#elif defined (SHADER_SPECULAR_NORMAL_UP)
    mediump vec3    specNormal = vec3(0.0, 1.0, 0.0);
#else
    mediump vec3    specNormal = varying_specNormal;
#endif
    
#if defined (SHADER_PER_PIXEL_SPECULAR)
    mediump	vec3    specEyeVec = normalize(varying_specPosWorld - uniform_eyePosition);;
    mediump	vec3    specRefVec = normalize(varying_specLightVec.xyz - specEyeVec);
    mediump	float	specCoeff = dot(specRefVec, normalize(specNormal.xyz));
    specCoeff = (pow(max(0.0, specCoeff), uniform_specularRGBPow.w) * varying_specLightVec.w);
#else
    // MAYA: This is how the Maya shader fx shader does specular...
    mediump	float   specCoeff = dot(normalize(varying_specRefVec.xyz), normalize(specNormal.xyz));
    specCoeff = (pow(max(0.0, specCoeff), uniform_specularRGBPow.w) * varying_specRefVec.w);
#endif

#if defined (SHADER_SPECULAR_ALPHA)
	//mediump	float	specCoeffTEMP = specCoeff;
    finalRGBA.xyz += (uniform_specularRGBPow.xyz * specCoeff * 0.25);
#endif

#if defined (SHADER_TEXTURE_SPECULAR)
	#if defined (SHADER_TEXTURE_ANIMUV) && defined (SHADER_SPECULAR_ALPHA)
	lowp vec4	specularTexel = TEXTURE2D(uniform_specularmap, varying_uv1);
	#else
	lowp vec4	specularTexel = TEXTURE2D(uniform_specularmap, varying_uv0);
	#endif
	specCoeff *= specularTexel.x;
#endif
	
#if defined (SHADER_SHADOW_RECEIVE)
	specCoeff *= shadow;
#endif

#if defined (SHADER_SPECULAR_ALPHA)
	specCoeff *= albedo.w;
#endif
	
    finalRGBA.xyz += (uniform_specularRGBPow.xyz * specCoeff);
#endif

    //--------------------------------------------------------------------------
    // Refraction...
#if defined (SHADER_TEXTURE_REFRACTION)
    lowp vec4	refractionTexel = TEXTURE2D(uniform_refractionmap, varying_refractionUV);
    finalRGBA.xyz = (finalRGBA.xyz * (1.0 - refractionTexel.w)) + (refractionTexel.xyz * refractionTexel.w);
#endif
    
	//--------------------------------------------------------------------------
	// Reflections...
	//--------------------------------------------------------------------------

#if defined (SHADER_TEXTURE_REFLECTION_SPHERE)
	mediump float reflectBlur = 0.0;
	
	// View-space spherical reflection map...
	mediump vec3 r = varying_reflectionUV.xyz;
	
	//	mediump float m = 2.0 * sqrt( pow( r.x, 2.0 ) + pow( r.y, 2.0 ) + pow( r.z + 1.0, 2.0 ) );
	mediump float rzplus1 = (r.z + 1.0);
	mediump float m = 2.0 * sqrt( (r.x * r.x) + (r.y * r.y) + (rzplus1 * rzplus1) );
	mediump vec2 rUV = r.xy / m + 0.5;
	lowp vec4	reflection	= TEXTURE2DBIAS(uniform_reflectionmap, rUV,  reflectBlur);

#if defined (SHADER_SHADOW_RECEIVE)
	reflection.xyz *= ((shadow * 0.5) + 0.5);
#endif
	finalRGBA.xyz += (reflection.xyz * uniform_reflectionColour.xyz);
	// Increase alpha by reflection intensity...
	finalRGBA.w = max(finalRGBA.w, dot(reflection.xyz, vec3(0.2989, 0.5870, 0.1140)));

#elif defined (SHADER_TEXTURE_REFLECTION_LATLONG)
	
	mediump vec3 r = normalize(varying_reflectionUV.xyz);
//	mediump vec2 rUV = (vec2((atan(r.x, r.z) + PI), acos(r.y)) / vec2(PI20, PI));		// XZ and Y Per pixel...
	mediump vec2 rUV = (vec2(((atan(r.x, r.z) + PI) / PI20), varying_reflectionUV.w));	// XZ per pixel and Y per vertex (y does not need to wrap)...
	mediump vec4 reflection	= TEXTURE2D(uniform_reflectionmap, rUV.xy);
#if defined (SHADER_LIGHT) && defined (SHADER_LIGHT_SPECULAR)
	#if defined (SHADER_TEXTURE_SPECULAR)
	reflection *= specularTexel.x;
	#endif
#endif
#if defined (SHADER_SHADOW_RECEIVE)
	reflection.xyz *= ((shadow * 0.5) + 0.5);
#endif
    //finalRGBA.xyz += (reflection.xyz * uniform_reflectionColour.xyz);
    mediump vec3 scaledReflectionRGB = (reflection.xyz * uniform_reflectionColour.xyz);
    finalRGBA.xyz += scaledReflectionRGB;
    // Increase alpha by reflection intensity...
    //finalRGBA.w = max(finalRGBA.w, dot(reflection.xyz, vec3(0.2989, 0.5870, 0.1140)));
    finalRGBA.w = min(1.0, (finalRGBA.w + max(max(scaledReflectionRGB.x, scaledReflectionRGB.y), scaledReflectionRGB.z)));

#elif defined (SHADER_TEXTURE_REFLECTION_PLANE)
	
 	#if defined (SHADER_MARSH_WATER)
		//mediump vec3	reflectWorld = normalize(reflect(eyeDirWorld, normalWorld)) * vec3(0.01, -1.0, 0.01) * distortionScale;
		mediump vec3	reflectWorld = normalize(reflect(eyeDirWorld, normalWorld)) * vec3(0.1, -1.0, 0.1) * distortionScale;
		
		// Prevent the reflection vector causing the reflection map to be sampled above the level of the reflection plane...
		reflectWorld.y = max(0.0, reflectWorld.y);
		highp vec4	posReflected = (varying_posWorld - vec4(reflectWorld, 0.0));
		highp vec4	posReflectedScreen = (uniform_viewProjectionMatrix * posReflected);
		highp vec2	refPlaneUV = ((posReflectedScreen.xy / posReflectedScreen.w) * vec2((0.5 * REFLECTION_SCREEN_X_SCALE), (0.5 * REFLECTION_SCREEN_Y_SCALE))) + 0.5;
		mediump vec4 reflection = TEXTURE2D(uniform_refractionmap, refPlaneUV);
		
		mediump float	light = max(0.0, dot(lightDirWorld, normalWorld));
		light = ((light * 0.4) + 0.8);
		reflection *= light;
		
//		finalRGBA.xyz = ((finalRGBA.xyz * reflection.xyz) * reflectionLevel) + (finalRGBA.xyz * (1.0 - reflectionLevel));
		finalRGBA.xyz = (reflection.xyz * reflectionLevel) + (finalRGBA.xyz * (1.0 - reflectionLevel));
    
        #if defined (SHADER_TEXTURE_OCEAN_OPACITY)
        finalRGBA.w = albedo.x;
        #else
        finalRGBA.w = 1.0;
        #endif
    #elif defined (SHADER_REFLECTIVE_SURFACE)
        // Simple reflection plane...
        mediump vec2 refPlaneUV = ((varying_posScreen.xy / varying_posScreen.w) * vec2((0.5 * REFLECTION_SCREEN_X_SCALE), (0.5 * REFLECTION_SCREEN_Y_SCALE))) + 0.5;
        lowp vec4 reflection = TEXTURE2D(uniform_refractionmap, refPlaneUV);

        mediump vec3 upNormal = vec3(0.0, 1.0, 0.0);
        mediump vec3 eyeToPos = vec3(varying_posWorld.xyz - uniform_eyePosition.xyz);
        mediump vec3 eyeDirWorld = normalize(eyeToPos);
        mediump float fresVal = fresnel(-eyeDirWorld, upNormal, 0.0, 1.0, 1.0);
    
        reflectionLevel *= fresVal;
    
        finalRGBA.xyz = (reflection.xyz * reflectionLevel) + (finalRGBA.xyz * (1.0 - reflectionLevel));
    #else
		// Simple reflection plane...
		mediump vec2 refPlaneUV = ((varying_posScreen.xy / varying_posScreen.w) * vec2((0.5 * REFLECTION_SCREEN_X_SCALE), (0.5 * REFLECTION_SCREEN_Y_SCALE))) + 0.5;
		lowp vec4 reflection = TEXTURE2D(uniform_refractionmap, refPlaneUV);
		finalRGBA.xyz *= reflection.xyz;
	#endif
	
#endif
	
	// Snow sparkle effect...
#if defined (SHADER_SNOW)
	mediump vec3	specNormalWorld = normalWorld0;
//  mediump vec3	snowLightDirWorld = normalize(uniform_lightDirection);
//  mediump vec3	snowLightDirWorld = normalize(uniform_horizonSunLightDirection);
    mediump vec3	snowLightDirWorld = normalize(uniform_specularDirection);

    mediump vec3 specRefVec	= normalize(snowLightDirWorld - eyeDirWorld);
	mediump	float specCoeff = clamp(dot(specRefVec, specNormalWorld), 0.0, 1.0);
	specCoeff = pow(specCoeff, uniform_specularRGBPow.w);
	specCoeff *= clamp(dot(specNormalWorld, snowLightDirWorld), 0.0, 1.0);
	
	specCoeff *= snowSparkleMask;

	// Scale the specular by the amount of deformation...
	specCoeff *= (varying_snowPosWorld_Deform.w * varying_snowPosWorld_Deform.w);
	
#if defined (SHADER_SHADOW_RECEIVE)
	specCoeff *= shadow;
#endif
	
    finalRGBA.xyz += (uniform_specularRGBPow.xyz * specCoeff);
#endif

//    //--------------------------------------------------------------------------
//    // Refraction...
//#if defined (SHADER_TEXTURE_REFRACTION)
//    lowp vec4	refractionTexel = TEXTURE2D(uniform_refractionmap, varying_refractionUV);
////    finalRGBA.xyz = refractionTexel.xyz;
//    finalRGBA.xyz = refractionTexel.xyz + finalRGBA.xyz;
//#endif
//
	//--------------------------------------------------------------------------

#if defined (SHADER_SHADOW_TRANSPARENT)
	finalRGBA = vec4(0.0, 0.0, 0.0, (((1.0 - shadow) * 0.5) * varying_rgba.w));
#endif

    //--------------------------------------------------------------------------
    
#if defined (SHADER_FOG)
	finalRGBA.xyz = (finalRGBA.xyz * varying_fogColour.w) + varying_fogColour.xyz;
#endif

    //--------------------------------------------------------------------------
    
#if defined (SHADER_HSV_ADJUST)
	mediump vec3 hsv = rgb2hsv(finalRGBA.xyz);
	hsv += uniform_hsvAdjust;
	finalRGBA.xyz = hsv2rgb(hsv);
#elif defined (SHADER_REMAP_HUE)
	finalRGBA.xyz = remapHue(finalRGBA.xyz, uniform_remapHue);
#endif

	//--------------------------------------------------------------------------

#if defined (SHADER_MAP_LOCKED_MASK) || defined (SHADER_MAP_LOCKED_CONSTANT)
#if defined (SHADER_MAP_LOCKED_MASK)
    // New tint which is a bit more blue with higher contrast...
    const lowp vec3 mapLockColourTint = vec3(0.1, 0.35, 0.55);
    const lowp vec3 mapLockColourAdd = vec3(0.11, 0.16, 0.22);
    const lowp vec3 greyDot = vec3(0.2989, 0.5870, 0.1140);
	mediump vec2 maskUV = (((varying_mapLockedMaskPosScreen.xy / varying_mapLockedMaskPosScreen.w) * vec2(0.5, 0.25)) + vec2(0.5, 0.75));
	mediump vec4 maskTexel = TEXTURE2D(uniform_refractionmap, maskUV.xy);

    lowp float greyScale = dot(finalRGBA.xyz, greyDot);
    lowp vec3 finalMapLockRGB = ((greyScale * mapLockColourTint) + mapLockColourAdd);

	finalRGBA.xyz = (finalMapLockRGB * maskTexel.x) + (finalRGBA.xyz * (1.0 - maskTexel.x));
#elif defined (SHADER_MAP_LOCKED_CONSTANT)
    //lowp float greyScale = dot(finalRGBA.xyz, greyDot);
    //finalRGBA.xyz = ((greyScale * mapLockColourTint) + mapLockColourAdd);

    // New simple tint for the world map...
//    const lowp vec3 finalMapLockRGB = vec3(0.10, 0.09, 0.37);
//    const lowp float lockRatio = 0.45;
    const lowp vec3 finalMapLockRGB = vec3(0.875, 0.796, 0.650);
#if defined (SHADER_SHADOW_RECEIVE)
    lowp float lockRatio = (0.30 + (shadow * 0.20));
#else
    lowp float lockRatio = 0.5;
#endif
    finalRGBA.xyz = (finalMapLockRGB * lockRatio) + (finalRGBA.xyz * (1.0 - lockRatio));
#endif
#endif
    
	//--------------------------------------------------------------------------

#if defined (SHADER_LIGHT) && defined (SHADER_LIGHT_SPECULAR)
    // Clamp to prevent black pixels caused by large alpha values on some GPU's...
    //LM: I've combined this clamp with a line from above to increase alpha by specular intensity...
    //finalRGBA.w = min(finalRGBA.w, 1.0);
    finalRGBA.w = min((finalRGBA.w + specCoeff), 1.0);
#endif

    //--------------------------------------------------------------------------

#if defined (SHADER_DISSOLVE)
    mediump vec4 dissolveTexel = texture2D(uniform_alphamap, varying_uv0);
    mediump float dissolveAlpha = step(varying_dissolve.x, dissolveTexel.x);
    if (dissolveAlpha <= 0.0)
    {
        discard;
    }
    mediump float dissolveSolid = step(varying_dissolve.y, dissolveTexel.x);
    mediump float dissolveMix = (dissolveAlpha - dissolveSolid);
    
    finalRGBA.xyz *= dissolveSolid;
    finalRGBA.xyz += (uniform_emissiveColour.xyz * dissolveMix);
    finalRGBA.w *= dissolveAlpha;
#endif

    //--------------------------------------------------------------------------
    
	FRAGCOLOUR = finalRGBA;

	//--------------------------------------------------------------------------

#if defined (SHADER_TEXTURE)
#if defined (SHADER_TEXTURE_SPECULAR)
#if defined (SHADER_SPECULAR_ALPHA)
//	FRAGCOLOUR = vec4(specularTexel.xyz, 1.0);
//	FRAGCOLOUR = vec4(specularTexel.xyz * albedo.w, 1.0);
//	FRAGCOLOUR = vec4(albedo.w, albedo.w, albedo.w, 1.0);
//	FRAGCOLOUR = vec4(specCoeff, specCoeff, specCoeff, 1.0);
//	FRAGCOLOUR = vec4(specCoeffTEMP, specCoeffTEMP, specCoeffTEMP, 1.0);
#endif
#endif
#endif
    
//#if defined (SHADER_SNOW)
//	FRAGCOLOUR = vec4(snowLighting, 1.0);
//#endif
    
//#if defined (SHADER_PER_PIXEL_SPECULAR)
//	FRAGCOLOUR = vec4(specEyeVec.xyz, 1.0);
//#endif

//#if defined (SHADER_LIGHT) || defined (SHADER_VCOL) || defined (SHADER_FIXED_COLOUR)
//    FRAGCOLOUR = varying_rgba;
//#endif

//#if defined (SHADER_SHADOW_RECEIVE) || defined (SHADER_MAP_CLOUD_SHADOWS) || defined (SHADER_SNOW) || defined (SHADER_TEXTURE_SHADOWMAP)
//    FRAGCOLOUR = vec4(varying_shadowAmbient.xyz, 1.0);
//#endif

	//--------------------------------------------------------------------------
	
#endif	//defined (SHADER_DEPTH_ONLY)
}

//------------------------------------------------------------------------------
