//
//  Shader_ES300_New.vsh
//  CityBuilder
//
//  Created by Lee Metcalfe on 23/06/2014.
//  Copyright (c) 2011 Midoki Ltd. All rights reserved.
//
//------------------------------------------------------------------------------

//==============================================================================
// GL specific defines...
//==============================================================================

#define SHADER_TYPE_VERTEX
//#define SHADER_TYPE_FRAGMENT

// Defines and macros to allow the shader code to be compiled for ES2 and ES3...
#if defined ES_300

#define ATTRIBUTE				in
#if defined (SHADER_TYPE_VERTEX)
#define VARYING                 out
#else
#define VARYING                 in
#endif
#define UNIFORM					uniform
#define FRAGCOLOUR				out_fragColour
#define SAMPLER2D				uniform sampler2D
#define SAMPLER2DSHADOW			uniform mediump sampler2DShadow
#define TEXTURE2D(a,b)			texture(a,b)
#define TEXTURE2DSHADOW(a,b)	texture(a,b.xyz/b.w)

#if defined (ANDROID)
#define TEXTURE2DBIAS(a,b,c)	texture(a,b)
#else
#define TEXTURE2DBIAS(a,b,c)	texture(a,b,c)
#endif

#else

#define ATTRIBUTE				attribute
#define VARYING                 varying
#define UNIFORM					uniform
#define FRAGCOLOUR				gl_FragColor
#define SAMPLER2D				uniform sampler2D
#define TEXTURE2D(a,b)			texture2D(a,b)

#if defined (OSX)
#define SAMPLER2DSHADOW			uniform sampler2DShadow
#define TEXTURE2DSHADOW(a,b)	shadow2DProj(a,b).x
#else
#define SAMPLER2DSHADOW			uniform sampler2D
#define TEXTURE2DSHADOW(a,b)	texture2DProj(a,b).x
#endif

#if defined (ANDROID)
#define TEXTURE2DBIAS(a,b,c)	texture2D(a,b)
#else
#define TEXTURE2DBIAS(a,b,c)	texture2D(a,b,c)
#endif

#endif

//------------------------------------------------------------------------------
//LM: Special case for devices that do not support vertex textures, undo the deformation defines but ensure that GRASS_USE_NORMAL_AS_XZ_OFFSET gets defined if required...

#if defined (SHADER_VERTEX_TEXTURE_UNSUPPORTED)
#if defined (SHADER_GRASS)
#if defined (SHADER_GRASS_DEFORMATION)
#define GRASS_USE_NORMAL_AS_XZ_OFFSET
#endif
#endif
#undef SHADER_GRASS_DEFORMATION
#undef SHADER_SNOW_DEFORMATION
#endif

//------------------------------------------------------------------------------

#define DECOMPRESSION_MULTIPLIER_POSITION       (1.0 / 32767.0)
#define DECOMPRESSION_MULTIPLIER_UV				(1.0 / 1024.0)
#define DECOMPRESSION_MULTIPLIER_WEIGHT			(1.0 / 255.0)

//------------------------------------------------------------------------------

//==============================================================================
// Common defines...
//==============================================================================

#define PI		3.1415926
#define PI20	(PI*2.0)
#define PI05	(PI*0.5)

//------------------------------------------------------------------------------
// Some special cases...
//------------------------------------------------------------------------------

#if defined (SHADER_MARSH_WATER)
#undef SHADER_TEXTURE_NORMAL
#undef SHADER_TEXTURE_NORMAL2
#undef SHADER_TEXTURE_ANIMUV
#undef SHADER_TEXTURE_REFLECTION_PLANE

#define SHADER_TEXTURE_NORMAL
#define SHADER_TEXTURE_NORMAL2
#define SHADER_TEXTURE_ANIMUV
#define SHADER_TEXTURE_REFLECTION_PLANE
#endif

//------------------------------------------------------------------------------

// If lighting is enabled, also force the fixed colour to be used...
#if defined (SHADER_LIGHT)
#define SHADER_FIXED_COLOUR
#endif

// Simplified define for the shadowmap case...
#if defined (SHADER_TEXTURE) && defined (SHADER_TEXTURE_MULTI) && defined (SHADER_TEXTURE_MULTI_SHADOWMAP)
#define SHADER_TEXTURE_SHADOWMAP
#endif

#define SHADER_LIGHT_POINT

//------------------------------------------------------------------------------

//==============================================================================
// Vertex shader defines...
//==============================================================================

// Legacy support for the classic 3 directional light plus ambient lighting solution...
//#if defined (SHADER_LIGHT) && !defined (SHADER_LIGHT_DIRECTIONAL)
//#define SHADER_LIGHT_CLASSIC
//#endif

// ...moving forward, use the new lighting solution for EVERYTHING!!!
#if defined (SHADER_LIGHT) && !defined (SHADER_LIGHT_CONSTANT)
#if defined (SHADER_VCOL_LIGHT_BAKE)
#define SHADER_LIGHT_CLASSIC
#else
#define SHADER_LIGHT_SH
#define SHADER_LIGHT_DIRECTIONAL
#endif
#endif

// Temporarily enable 4 bones per vertex for all skin shaders...
#undef SHADER_SKIN_4BONE
#define SHADER_SKIN_4BONE

//------------------------------------------------------------------------------

//==============================================================================
// Uniforms...
//==============================================================================

UNIFORM highp	mat4 uniform_viewProjectionMatrix;

#if !defined (SHADER_SKIN)
UNIFORM highp	mat4 uniform_modelMatrix;
#endif

#if defined (SHADER_LIGHT)
#if defined (SHADER_LIGHT_CLASSIC)
UNIFORM highp	mat4 uniform_lightDirectionMat;
UNIFORM highp	mat4 uniform_lightColourMat;
#endif
#if defined (SHADER_LIGHT_DIRECTIONAL)
UNIFORM highp	vec3 uniform_lightDirection;
UNIFORM highp	vec3 uniform_lightColour;
#endif
#if defined (SHADER_LIGHT_AMBIENT)
UNIFORM highp	vec3 uniform_lightAmbient;
#endif
#if defined (SHADER_LIGHT_SPECULAR)
UNIFORM mediump vec4 uniform_specularRGBPow;
UNIFORM mediump vec3 uniform_specularDirection;
#endif
#else
#if defined (SHADER_LIGHT_CONSTANT)
UNIFORM highp	vec3 uniform_lightAmbient;
UNIFORM highp	vec3 uniform_lightColour;
#endif
#endif

#if defined (SHADER_LIGHT_POINT)
UNIFORM highp	vec3 uniform_pointLightPosition;
UNIFORM highp	vec3 uniform_pointLightColour;
UNIFORM highp	vec3 uniform_pointLightAttenuation;
#endif

#if defined (SHADER_FIXED_COLOUR)
UNIFORM highp	vec4 uniform_colour;
#endif

#if defined (SHADER_TEXTURE_ANIMUV) || defined (SHADER_WATER_FOAM) || defined (SHADER_MAP_CLOUD_SHADOWS) || defined (SHADER_MARSH_WATER)
UNIFORM highp	vec4 uniform_animUV;
#endif

#if defined (SHADER_TEXTURE_ANIMUVFRAME) || defined (SHADER_TEXTURE_ANIMUVFRAMEBLEND)
UNIFORM highp	vec4 uniform_animUVFrame;
#endif

#if defined (SHADER_TEXTURE_UVOFFSET)
UNIFORM highp	vec2 uniform_uvOffset;
#endif

#if defined (SHADER_SHADOW_RECEIVE)
UNIFORM highp	mat4 uniform_shadowProjectionMatrix;
UNIFORM mediump	float uniform_shadowLightingModifier;
#endif

#if defined (SHADER_FOG)
#define SHADER_FOG_EXP2
UNIFORM mediump	vec4 uniform_fogPlaneY;
UNIFORM mediump	vec4 uniform_fogPlaneZ;
UNIFORM mediump vec4 uniform_fogColour;
#endif

#if defined (SHADER_TEXTURE_REFLECTION_SPHERE) || defined (SHADER_TEXTURE_REFLECTION_LATLONG) || defined (SHADER_LIGHT_SPECULAR) || defined (SHADER_FOG_SUN_HAZE) || (defined (SHADER_SNOW) && defined (SHADER_SNOW_DEFORMATION)) || defined (SHADER_OUTLINE)
UNIFORM highp	vec3 uniform_eyePosition;
#endif

#if defined (SHADER_FOG_SUN_HAZE)
UNIFORM highp	vec4 uniform_sunHazeColourFog;
UNIFORM mediump vec3 uniform_horizonSunLightDirection;
#endif

#if defined (SHADER_GRASS)
#if defined (SHADER_GRASS_DEFORMATION)
#define GRASS_USE_NORMAL_AS_XZ_OFFSET
SAMPLER2D uniform_deformmap;
UNIFORM mediump	vec4 uniform_grassDeformation;
#endif
UNIFORM mediump	vec4 uniform_grassSettings;
#endif

#if defined (SHADER_SNOW)
#if defined (SHADER_SNOW_DEFORMATION)
SAMPLER2D uniform_deformmap;
UNIFORM mediump	vec4 uniform_grassDeformation;
UNIFORM mediump vec3 uniform_snowSettings;
#endif
#endif

#if defined (SHADER_OUTLINE)
UNIFORM mediump vec3 uniform_outlineScale;
#endif

#if defined (SHADER_TEXTURE_REFRACTION)
UNIFORM mediump	vec4 uniform_refractUVScaleOffset;
#endif

#if defined (SHADER_MARSH_WATER)
UNIFORM highp vec4 uniform_oceanWaterSettings;
#endif

#if defined (SHADER_DISSOLVE)
UNIFORM mediump float uniform_animationValue;
UNIFORM mediump vec4 uniform_emissiveColour;
#endif

//------------------------------------------------------------------------------

#if defined (SHADER_SKIN)
#if defined (ANDROID)
UNIFORM highp	vec4 uniform_boneMat[MAX_BONES_PER_SKIN*4];	//LM: Some Android devices cannot index matrix arrays by variable...
#else
UNIFORM highp	mat4 uniform_boneMat[MAX_BONES_PER_SKIN];
#endif
#endif

//------------------------------------------------------------------------------

#if defined (SHADER_LIGHT_SH)
UNIFORM mediump	vec4 uniform_SHCoeff[7];
#endif

//==============================================================================
// Vertex shader inputs...
//==============================================================================

ATTRIBUTE highp	vec4 attrib_position;

#if defined (SHADER_LIGHT) || defined (SHADER_FRESNEL_TRANS) //|| defined (SHADER_OUTLINE)
ATTRIBUTE highp	vec3 attrib_normal;
#endif

#if defined (SHADER_VCOL) || defined (SHADER_GRASS) || defined (SHADER_OUTLINE) || defined (SHADER_MARSH_WATER) || defined (SHADER_FADE_FOG_BY_ALPHA) || defined (SHADER_TEXTURE_MULTI_VRED)
ATTRIBUTE highp	vec4 attrib_rgba;
#endif

#if defined (SHADER_TEXTURE) || defined (SHADER_TEXTURE_EMISSIVE) || defined (SHADER_SNOW_NORMALMAP)
ATTRIBUTE highp	vec2 attrib_uv0;
#if defined (SHADER_TEXTURE_MULTI_UV2) || defined (SHADER_MARSH_WATER)
ATTRIBUTE highp	vec2 attrib_uv1;
#endif
#endif

#if defined (SHADER_SKIN)
ATTRIBUTE highp	vec4 attrib_boneWeights;
ATTRIBUTE highp	vec4 attrib_boneIndices;
#endif

//------------------------------------------------------------------------------

//==============================================================================
// Vertex shader outputs...
//==============================================================================

#if defined (SHADER_LIGHT) || defined (SHADER_VCOL) || defined (SHADER_FIXED_COLOUR) || defined (SHADER_TEXTURE_MULTI_VRED)
VARYING lowp    vec4 varying_rgba;
#endif

//------------------------------------------------------------------------------
// UV's - the conditions here match the code in the vertex shader, this is required so we can vary precision for the animated uv's...

#if defined (SHADER_TEXTURE) || defined (SHADER_TEXTURE_EMISSIVE) || defined (SHADER_SNOW_NORMALMAP)
#if defined (SHADER_WATER_FOAM)

VARYING lowp	vec2 varying_uv0;
VARYING mediump	vec4 varying_uv1uv2;

#elif defined (SHADER_MARSH_WATER)

VARYING mediump	vec2 varying_uv0;
VARYING mediump	vec4 varying_uv1uv2;

#elif defined (SHADER_TEXTURE_ANIMUV)

// Animated (scrolling) uv's...
VARYING mediump	vec2 varying_uv0;
#if defined (SHADER_TEXTURE_MULTI_UV2)
VARYING mediump	vec2 varying_uv1;
#endif
#if defined (SHADER_SPECULAR_ALPHA)
VARYING lowp	vec2 varying_uv1;
#endif

#elif defined (SHADER_TEXTURE_ANIMUVFRAME) || defined (SHADER_TEXTURE_ANIMUVFRAMEBLEND)

// Animated (frame cycling) uv's...
VARYING mediump vec2 varying_uv0;
#if defined (SHADER_TEXTURE_ANIMUVFRAMEBLEND)
VARYING highp   vec3 varying_animUVFrameBlend;
#endif

#elif defined (SHADER_TEXTURE_EMISSIVE) || defined (SHADER_SNOW_NORMALMAP)

VARYING lowp    vec2 varying_uv0;

#else

// Static uv's...
#if defined (SHADER_TEXTURE)
VARYING lowp	vec2 varying_uv0;
#if defined (SHADER_TEXTURE_MULTI_UV2)
VARYING lowp	vec2 varying_uv1;
#endif
#endif

#endif  //defined (SHADER_WATER_FOAM)
#endif  //defined (SHADER_TEXTURE) || defined (SHADER_TEXTURE_EMISSIVE)

//------------------------------------------------------------------------------

#if defined (SHADER_TEXTURE_REFLECTION_SPHERE) || defined (SHADER_TEXTURE_REFLECTION_LATLONG)
VARYING mediump vec4 varying_reflectionUV;
#endif

#if defined (SHADER_TEXTURE_REFRACTION)
VARYING mediump vec2 varying_refractionUV;
#endif

#if defined (SHADER_MAP_CLOUD_SHADOWS)
VARYING lowp	vec4 varying_cloud_uv0uv1;
#endif

//------------------------------------------------------------------------------

#if defined (SHADER_LIGHT) && defined (SHADER_LIGHT_SPECULAR)
////VARYING mediump	vec3 varying_specLightVec;
////VARYING mediump	vec3 varying_specNormalVec;
////VARYING mediump	vec3 varying_specEyeVec;
//VARYING mediump	vec3 varying_specRefVec;
//VARYING mediump	vec3 varying_specEyeVec;
#if defined (SHADER_PER_PIXEL_SPECULAR)
VARYING mediump	vec4 varying_specLightVec;
VARYING highp	vec3 varying_specPosWorld;
#else
VARYING mediump	vec4 varying_specRefVec;
#endif
#if !defined (SHADER_SPECULAR_NORMAL_UP)
VARYING mediump	vec3 varying_specNormal;
#endif
#endif

#if defined (SHADER_SHADOW_RECEIVE) || defined (SHADER_MAP_CLOUD_SHADOWS) || defined (SHADER_SNOW) || defined (SHADER_TEXTURE_SHADOWMAP)
VARYING lowp    vec3 varying_shadowAmbient;
#endif
#if defined (SHADER_SHADOW_RECEIVE)
VARYING highp	vec4 varying_shadowUVZW;
#endif

#if defined (SHADER_FOG)
VARYING mediump	vec4 varying_fogColour;
#endif

#if defined (SHADER_TEXTURE_REFLECTION_PLANE) //|| (defined (SHADER_SHADOW_RECEIVE) && defined (SHADER_SHADOW_DITHER))
VARYING highp vec4 varying_posScreen;
#endif

#if defined (SHADER_MARSH_WATER) || defined (SHADER_REFLECTIVE_SURFACE) || defined (SHADER_FRESNEL_TRANS)
VARYING highp vec4 varying_posWorld;
#endif

#if defined (SHADER_SNOW)
VARYING mediump	vec3 varying_snowNormal;
VARYING highp	vec4 varying_snowPosWorld_Deform;
#endif

#if defined (SHADER_MAP_LOCKED_MASK)
VARYING highp   vec4 varying_mapLockedMaskPosScreen;
#endif

#if defined (SHADER_DISSOLVE)
VARYING mediump vec2 varying_dissolve;
#endif

#if defined (SHADER_FRESNEL_TRANS)
VARYING mediump vec3 varying_fresnelTransNormal;
#endif

//------------------------------------------------------------------------------

//------------------------------------------------------------------------------

#if defined (SHADER_LIGHT_CLASSIC)
mediump vec4 calculateClassicLightingDots(vec4 n, mat4 ldirmat)
{
	mediump vec4	light_dots = (ldirmat * n);
	light_dots		= max(light_dots, 0.0);
	light_dots.w	= 1.0;
	
	return light_dots;
}
#endif

//------------------------------------------------------------------------------

#if defined (SHADER_LIGHT_POINT)
mediump vec3 calculatePointLight(mediump vec3 pos, mediump vec3 n, mediump vec3 lpos, mediump vec3 latt, mediump vec3 lcol)
{
	mediump vec3	pl2pos = (pos - lpos);
	mediump float	pl2posDist = max(length(pl2pos), 0.001);
	mediump vec3	pl2posN = (pl2pos / pl2posDist);
	mediump float	plDot = max(0.0, dot(-pl2posN, n));
	mediump float	pl2posDistQuadratic = (pl2posDist * pl2posDist);
	
//	mediump float	plAtt = (1.0 / (latt.x + (latt.y * pl2posDist) + (latt.z * pl2posDistQuadratic)));

	// Modified attenuation, rather than constant/linear/quadratic we actually use linear/quadratic/cubic to fully support Maya style point lights...
	mediump float	pl2posDistCubic = (pl2posDistQuadratic * pl2posDist);
	mediump float	plAtt = max(0.0, (1.0 / dot(latt, vec3(pl2posDist, pl2posDistQuadratic, pl2posDistCubic))));

	mediump vec3	result = (lcol * (plDot * plAtt));

	return result;
}
#endif

//------------------------------------------------------------------------------

#if defined (SHADER_LIGHT_SH)
mediump vec3 calculateSphericalHarmonics(mediump vec3 n)
{
//	const mediump float C1 = 0.429043;
//	const mediump float C2 = 0.511664;
//	const mediump float C3 = 0.743125;
//	const mediump float C4 = 0.886227;
//	const mediump float C5 = 0.247708;

	mediump vec3	flip	= (n.xzy * n.zyx);
	mediump vec3	square	= (n * n);

//	mediump vec3	result  = C4 * uniform_SHCoeff[0].xyz;
//					result += C2 * 2.0 * uniform_SHCoeff[1].xyz * n.x;
//					result += C2 * 2.0 * uniform_SHCoeff[2].xyz * n.y;
//					result += C2 * 2.0 * uniform_SHCoeff[3].xyz * n.z;
//					result += C1 * 2.0 * uniform_SHCoeff[4].xyz * flip.x;
//					result += C1 * 2.0 * uniform_SHCoeff[5].xyz * flip.y;
//					result += C1 * 2.0 * uniform_SHCoeff[6].xyz * flip.z;
//					result += C3 * uniform_SHCoeff[7].xyz * (square.z - C5);
//					result += C1 * uniform_SHCoeff[8].xyz * (square.x - square.y);

//	// We now roll most of the constants above into the coefficient uniforms...
//	mediump vec3	result  = uniform_SHCoeff[0].xyz;
//					result += uniform_SHCoeff[1].xyz * n.x;
//					result += uniform_SHCoeff[2].xyz * n.y;
//					result += uniform_SHCoeff[3].xyz * n.z;
//					result += uniform_SHCoeff[4].xyz * flip.x;
//					result += uniform_SHCoeff[5].xyz * flip.y;
//					result += uniform_SHCoeff[6].xyz * flip.z;
//					result += uniform_SHCoeff[7].xyz * (square.z - C5);
//					result += uniform_SHCoeff[8].xyz * (square.x - square.y);

	// We now roll most of the constants above into the coefficient uniforms, also now compressed into fewer vectors...
	mediump vec3	result  = uniform_SHCoeff[0].xyz;
					result += uniform_SHCoeff[1].xyz * n.x;
					result += uniform_SHCoeff[2].xyz * n.y;
					result += uniform_SHCoeff[3].xyz * n.z;
					result += uniform_SHCoeff[4].xyz * flip.x;
					result += uniform_SHCoeff[5].xyz * flip.y;
					result += uniform_SHCoeff[6].xyz * flip.z;
					result += vec3(uniform_SHCoeff[0].w, uniform_SHCoeff[1].w, uniform_SHCoeff[2].w) * (square.z - uniform_SHCoeff[6].w);
					result += vec3(uniform_SHCoeff[3].w, uniform_SHCoeff[4].w, uniform_SHCoeff[5].w) * (square.x - square.y);
	
	result = max(result, 0.0);

	return result;
}
#endif

//------------------------------------------------------------------------------

#if defined (SHADER_FOG)
mediump float calculateFogValue(mediump vec4 posModel, mediump vec4 fogPlaneY, mediump vec4 fogPlaneZ, mediump float fogDensity)
{
	mediump float	fogValue	= max(dot(fogPlaneY, posModel), dot(fogPlaneZ, posModel));
					fogValue	= max(fogValue, 0.0);

#if defined (SHADER_FOG_EXP2)
	#define LOG2 1.442695
	mediump float	result		= exp2(-(fogValue*fogValue*LOG2));
					result		= clamp(result, 0.0, 1.0);
					result		= (1.0 - result) * fogDensity;
#else
	mediump float	result		= min(fogValue, 1.0) * fogDensity;
#endif

	return result;
}
#endif

//------------------------------------------------------------------------------

#if defined (SHADER_FOG)
mediump vec4 calculateFogColour(mediump vec3 fogColour, mediump float fogValue)
{
	mediump vec4	result = vec4((uniform_fogColour.xyz * fogValue), (1.0 - fogValue));
	
	return result;
}
#endif

//------------------------------------------------------------------------------

#if defined (SHADER_FOG) && defined (SHADER_FOG_SUN_HAZE) && defined (SHADER_LIGHT)
mediump vec4 calculateFogColourWithSunHaze(mediump vec3 fogColour, mediump float fogValue, mediump vec4 sunHazeFogColour, mediump vec3 eye, mediump vec3 light)
{
	mediump float	hazeFactor	= max(0.0, dot(light, eye));
					hazeFactor	= (hazeFactor * hazeFactor) * sunHazeFogColour.w;//pow(hazeFactor, 8.0);
	mediump vec3	hazeColour	= (sunHazeFogColour.xyz * hazeFactor) + (fogColour.xyz * (1.0 - hazeFactor));
	
	mediump vec4	result		= vec4((hazeColour.xyz * fogValue), (1.0 - fogValue));
	
	return result;
}
#endif

//------------------------------------------------------------------------------

void main()
{
	//--------------------------------------------------------------------------
	// Position...
	//--------------------------------------------------------------------------
	
    highp vec3	posScaled	= (attrib_position.xyz * (attrib_position.w * DECOMPRESSION_MULTIPLIER_POSITION));
    highp vec4	posScaled4	= vec4(posScaled, 1.0);

#if defined (SHADER_SKIN)
	highp vec4	weights		= (attrib_boneWeights * DECOMPRESSION_MULTIPLIER_WEIGHT);
#if defined (ANDROID)
    // Some Android devices cannot index matrix arrays by variable...
	int			bone0Index	= int(attrib_boneIndices.x) * 4;
	int			bone1Index	= int(attrib_boneIndices.y) * 4;
	highp mat4	bone0Mat3 = mat4(uniform_boneMat[bone0Index+0], uniform_boneMat[bone0Index+1], uniform_boneMat[bone0Index+2], uniform_boneMat[bone0Index+3]);
	highp mat4	bone1Mat3 = mat4(uniform_boneMat[bone1Index+0], uniform_boneMat[bone1Index+1], uniform_boneMat[bone1Index+2], uniform_boneMat[bone1Index+3]);
#if defined (SHADER_SKIN_4BONE)
	int			bone2Index	= int(attrib_boneIndices.z) * 4;
	int			bone3Index	= int(attrib_boneIndices.w) * 4;
	highp mat4	bone2Mat3 = mat4(uniform_boneMat[bone2Index+0], uniform_boneMat[bone2Index+1], uniform_boneMat[bone2Index+2], uniform_boneMat[bone2Index+3]);
	highp mat4	bone3Mat3 = mat4(uniform_boneMat[bone3Index+0], uniform_boneMat[bone3Index+1], uniform_boneMat[bone3Index+2], uniform_boneMat[bone3Index+3]);
#endif
#else
	highp mat4	bone0Mat3	= uniform_boneMat[int(attrib_boneIndices.x)];
	highp mat4	bone1Mat3	= uniform_boneMat[int(attrib_boneIndices.y)];
#if defined (SHADER_SKIN_4BONE)
	highp mat4	bone2Mat3	= uniform_boneMat[int(attrib_boneIndices.z)];
	highp mat4	bone3Mat3	= uniform_boneMat[int(attrib_boneIndices.w)];
#endif
#endif
	highp vec4	posSkin		= ((bone0Mat3 * posScaled4) * weights.x);
				posSkin    += ((bone1Mat3 * posScaled4) * weights.y);
#if defined (SHADER_SKIN_4BONE)
				posSkin    += ((bone2Mat3 * posScaled4) * weights.z);
				posSkin    += ((bone3Mat3 * posScaled4) * weights.w);
#endif
				posScaled4	= posSkin;
	highp vec4	posWorld	= posSkin;
#else
	highp vec4	posWorld	= (uniform_modelMatrix * posScaled4);
#endif //defined (SHADER_SKIN)
	
	//--------------------------------------------------------------------------
	// Grass...
	//--------------------------------------------------------------------------

#if defined (SHADER_GRASS)
#if defined (SHADER_GRASS_TREE_SWAY)
	//	mediump float grassBendY = posScaled4.y * 0.2;
	mediump float grassBendY = attrib_rgba.y;
	mediump float bendFactor = (grassBendY * grassBendY);
#else
	mediump float grassBendY = (attrib_rgba.y * 0.88);
	mediump float bendFactor = min((grassBendY * 0.25), ((grassBendY * grassBendY) * 0.4));
#endif
	mediump vec2 grassXZOffset = (uniform_grassSettings.xy * bendFactor);
	mediump float grassYScale = 1.0;
	
	// Length correct for wind bend...
#if defined (SHADER_GRASS_TREE_SWAY)
	//	grassYScale -= (length(grassXZOffset) * 0.25);
	// New y coordinate modification that creates a more realistic 'bend' effect rather than a 'skew'...
	grassYScale -= (length(grassXZOffset) * 0.2);
	grassYScale -= dot(uniform_grassSettings.zw, posScaled4.xz) * grassBendY;//bendFactor;//0.6;
#else
	grassYScale -= (length(grassXZOffset) * 0.6);
#endif
	
	// Apply deformation map...
#if defined (SHADER_GRASS_DEFORMATION)
	mediump vec2 deformUV = ((posWorld.xz - uniform_grassDeformation.xy) * uniform_grassDeformation.zw);
	deformUV.x = -deformUV.x;
	deformUV = ((deformUV * 0.5) + 0.5);
	mediump vec4 deformTexel = TEXTURE2D(uniform_deformmap, deformUV);
	
	mediump vec2 forceVector = ((deformTexel.xy - 0.5) * 2.0) * vec2(1.0, -1.0);
	mediump float forceIntensity = deformTexel.z;
	
	// This tweak helps make the grass recover slower and more smoothly...
	forceIntensity = (1.0 - forceIntensity);
	forceIntensity *= forceIntensity;
	forceIntensity *= forceIntensity;
	forceIntensity = (1.0 - forceIntensity);
	
	mediump vec2 forcePosOffset = (forceVector * grassBendY);
	mediump float forceYScale = (1.0 - (forceIntensity * 0.95));
	
	// Lerp away the force effect...
	forcePosOffset *= forceIntensity;
	forceYScale = ((forceYScale * forceIntensity) + (1.0 - forceIntensity));
	
	mediump float forceEffect = forceIntensity;
	grassXZOffset = (forcePosOffset * forceEffect) + (grassXZOffset * (1.0 - forceEffect));
	grassYScale = (forceYScale * forceEffect) + (grassYScale * (1.0 - forceEffect));
	
	//EXPERIMENTAL
	//mediump float windGustIntensity = (1.0 - deformTexel.z);
	//mediump float windGustIntensity = max(0.0, (1.0 - (deformTexel.z * 8.0)));
	mediump float windGustIntensity = max(0.0, (1.0 - (deformTexel.z * 2.0)));
	grassXZOffset += (windGustIntensity * (forcePosOffset * 0.5));
	
	// Use the alpha channel of the deform map to give us finer control over the grass...
	grassXZOffset *= deformTexel.w;
	grassYScale *= deformTexel.w;
    
    // Reduce lateral movement relative to deformation...
    grassXZOffset *= grassYScale;
#endif
	
#if defined (GRASS_USE_NORMAL_AS_XZ_OFFSET)
	mediump vec4 posWorldDiff = vec4(attrib_normal.x, 0.0, attrib_normal.z, 0.0);
#if defined (SHADER_GRASS_DEFORMATION)
	posWorld += (uniform_modelMatrix * posWorldDiff) * deformTexel.w;
#else
	posWorld += (uniform_modelMatrix * posWorldDiff);
#endif
#endif
	
	// Modify the world space position...
	posWorld.xz += grassXZOffset;
#if defined (SHADER_GRASS_TREE_SWAY)
	posWorld.y -= (1.0 - grassYScale);
#elif defined (SHADER_GRASS_DEFORMATION)
	posWorld.y -= (posScaled4.y * (1.0 - grassYScale));
#else
	posWorld.y -= (grassBendY * (1.0 - grassYScale));
#endif
	
	// Modify the model space normal...
    mediump vec3 normalGrass = attrib_normal.xyz;
    
	//	mediump float grassNormalModifier = sign(grassBendY);
	mediump float grassNormalModifier = min(1.0, (bendFactor * 8.0));
	
	normalGrass.xz -= (uniform_grassSettings.zw * 0.075 * grassNormalModifier);
#if defined (SHADER_GRASS_DEFORMATION)
	normalGrass.xz += (uniform_grassSettings.zw * (forceVector * 0.25) * grassNormalModifier);
#endif
	normalGrass = normalize(normalGrass);
#endif
	
#if defined (SHADER_SNOW)
#if defined (SHADER_SNOW_DEFORMATION)
	mediump float	snowDeformDepth = uniform_snowSettings.y;
	
	mediump vec2 deformUV = ((posWorld.xz - uniform_grassDeformation.xy) * uniform_grassDeformation.zw);
	deformUV.x = -deformUV.x;
	deformUV = ((deformUV * 0.5) + 0.5);
	mediump vec4 deformTexel = TEXTURE2D(uniform_deformmap, deformUV);
	
	posWorld.y -= (deformTexel.z * snowDeformDepth);
	
	mediump vec3 normalSnow = attrib_normal.xyz;
	normalSnow.xz += ((deformTexel.xy * 2.0) - 1.0) * vec2(-1.0, 1.0);
	normalSnow = normalize(normalSnow);
	
	//	mediump float snowLightDist = length(posWorld.xyz - uniform_eyePosition.xyz) * 1.1;
//	mediump vec3 snowEyePos = uniform_eyePosition.xyz;
//	mediump vec3 eyeToPos = (posWorld.xyz - snowEyePos.xyz);
//	mediump float snowLightDist = dot(eyeToPos, uniform_eyeDirection.xyz);
	
	varying_snowPosWorld_Deform = vec4(posWorld.xyz, (1.0 - deformTexel.z));
#else
	varying_snowPosWorld_Deform = vec4(posWorld.xyz, 1.0);
#endif
#endif

#if defined (SHADER_MARSH_WATER) || defined (SHADER_REFLECTIVE_SURFACE) || defined (SHADER_FRESNEL_TRANS)
	varying_posWorld = posWorld;
#endif

	highp vec4 posScreen = (uniform_viewProjectionMatrix * posWorld);
	
#if defined (SHADER_TEXTURE_REFLECTION_PLANE) //|| (defined (SHADER_SHADOW_RECEIVE) && defined (SHADER_SHADOW_DITHER))
	varying_posScreen = posScreen;
#endif

#if defined (SHADER_MAP_LOCKED_MASK)
	highp vec4	mapLockedMaskPosWorld = vec4(posWorld.x, 0.0, posWorld.z, 1.0);
	varying_mapLockedMaskPosScreen = (uniform_viewProjectionMatrix * mapLockedMaskPosWorld);
#endif
	
	//--------------------------------------------------------------------------
	// Normal...
	//--------------------------------------------------------------------------
    // Note: This needs to be here so we can use the modified grass/snow normals if required...

#if defined (SHADER_LIGHT) || defined (SHADER_FRESNEL_TRANS) //|| defined (SHADER_OUTLINE)
#if defined (SHADER_SKIN)
	vec3	normalSkin	= ((mat3(bone0Mat3) * attrib_normal) * weights.x);
            normalSkin += ((mat3(bone1Mat3) * attrib_normal) * weights.y);
#if defined (SHADER_SKIN_4BONE)
            normalSkin += ((mat3(bone2Mat3) * attrib_normal) * weights.z);
            normalSkin += ((mat3(bone3Mat3) * attrib_normal) * weights.w);
#endif
	// Normalize the skinned normal to prevent scale in the bone matrices affecting the lighting...
	mediump vec3	normalWorld	= normalize(normalSkin.xyz);
#elif defined (SHADER_GRASS)
#if defined (SHADER_GRASS_TREE_SWAY)
    mediump vec3    normalWorld	= normalize(mat3(uniform_modelMatrix) * normalGrass.xyz);
#else
	mediump vec3	normalWorld	= normalGrass.xyz;
#endif
#elif defined (SHADER_SNOW_DEFORMATION)
	mediump vec3	normalWorld = normalSnow.xyz;
#else
	mediump vec3	normalWorld	= normalize(mat3(uniform_modelMatrix) * attrib_normal.xyz);
#endif
#endif
	
#if defined (SHADER_SNOW)
	varying_snowNormal = normalWorld;
#endif
    
#if defined (SHADER_FRESNEL_TRANS)
    varying_fresnelTransNormal = normalWorld;
#endif

#if defined (SHADER_OUTLINE)
    mediump vec3  attrib_outline_normal = normalize((attrib_rgba.xyz - 0.5) * 2.0);
#if defined (SHADER_SKIN)
    mediump vec3	outlineNormalSkin	= ((mat3(bone0Mat3) * attrib_outline_normal) * weights.x);
    outlineNormalSkin += ((mat3(bone1Mat3) * attrib_outline_normal) * weights.y);
#if defined (SHADER_SKIN_4BONE)
    outlineNormalSkin += ((mat3(bone2Mat3) * attrib_outline_normal) * weights.z);
    outlineNormalSkin += ((mat3(bone3Mat3) * attrib_outline_normal) * weights.w);
#endif
    mediump vec3 outlineNormalWorld	= normalize(outlineNormalSkin.xyz);
#else
    mediump vec3 outlineNormalWorld	= normalize(mat3(uniform_modelMatrix) * attrib_outline_normal);
#endif
    mediump vec4 outlineNormalScreen = (uniform_viewProjectionMatrix * vec4(outlineNormalWorld, 0.0));
    
    // This will guard against a zero length xy vector which would cause a NAN when normalised...
    outlineNormalScreen.y += (1.0 - sign(dot(outlineNormalScreen.xy, outlineNormalScreen.xy)));
    
    posScreen.xy += (normalize(outlineNormalScreen.xy) * uniform_outlineScale.xy * posScreen.w);

    // Calculate a biased z to (hopefully) match the xy offset we've applied...
//  highp vec4 outlinePosScreen = (posWorld + (vec4(outlineNormalWorld, 0.0) * uniform_outlineScale.z));
    mediump vec3 outlinePosToEye = normalize(uniform_eyePosition.xyz - posWorld.xyz);
    highp vec4 outlinePosScreen = (posWorld + (vec4(outlinePosToEye, 0.0) * uniform_outlineScale.z));
    outlinePosScreen = (uniform_viewProjectionMatrix * outlinePosScreen);
    posScreen.z = ((outlinePosScreen.z / outlinePosScreen.w) * posScreen.w);
#endif
    
    gl_Position = posScreen;

	//--------------------------------------------------------------------------
	// Texture coordinates...
	//--------------------------------------------------------------------------

#if defined (SHADER_TEXTURE) || defined (SHADER_TEXTURE_EMISSIVE) || defined (SHADER_SNOW_NORMALMAP)
	
#if defined (SHADER_WATER_FOAM)
	
	mediump	vec2	baseUVOffset1 = fract(uniform_animUV.xy * (uniform_animUV.zw * -0.5));
	mediump	vec2	baseUVOffset2 = fract(uniform_animUV.xy * uniform_animUV.zw);
	varying_uv0			= (attrib_uv0 * DECOMPRESSION_MULTIPLIER_UV);
	varying_uv1uv2.xy   = (varying_uv0 + baseUVOffset1) + vec2(0.0, 0.25);
	varying_uv1uv2.zw   = (varying_uv0 + baseUVOffset2) + vec2(0.0, 0.50);
	
#elif defined (SHADER_MARSH_WATER)
	
	mediump	vec2	baseUVOffset1 = fract(uniform_animUV.xy * uniform_animUV.zw);
    mediump	vec2	baseUVOffset2 = fract(uniform_animUV.xy * uniform_oceanWaterSettings.xy);

    varying_uv0			= (attrib_uv0 * DECOMPRESSION_MULTIPLIER_UV);
    varying_uv1uv2.xy	= (attrib_uv1 * DECOMPRESSION_MULTIPLIER_UV) + baseUVOffset1;
    varying_uv1uv2.zw	= (attrib_uv1 * DECOMPRESSION_MULTIPLIER_UV) + baseUVOffset2;

#elif defined (SHADER_TEXTURE_ANIMUV)
	// Animated (scrolling) uv's...
	mediump	vec2	baseUVOffset = fract(uniform_animUV.xy * uniform_animUV.zw);
	varying_uv0			= ((attrib_uv0 * DECOMPRESSION_MULTIPLIER_UV) + baseUVOffset);
#if defined (SHADER_TEXTURE_MULTI_UV2)
	varying_uv1			= ((attrib_uv1 * DECOMPRESSION_MULTIPLIER_UV) + baseUVOffset);
#endif
#if defined (SHADER_SPECULAR_ALPHA)
	varying_uv1			= (attrib_uv0 * DECOMPRESSION_MULTIPLIER_UV);
#endif
	
#elif defined (SHADER_TEXTURE_ANIMUVFRAME) || defined (SHADER_TEXTURE_ANIMUVFRAMEBLEND)
	// Animated (frame cycling) uv's...
	mediump vec2 baseUV = (attrib_uv0 * DECOMPRESSION_MULTIPLIER_UV);
	varying_uv0			= (baseUV + uniform_animUVFrame.xy);
#if defined (SHADER_TEXTURE_ANIMUVFRAMEBLEND)
	varying_animUVFrameBlend.xy	= (baseUV + (floor(uniform_animUVFrame.zw) * (1.0 / 32768.0)));
	varying_animUVFrameBlend.z	= fract(uniform_animUVFrame.z) * 2.0;
#endif
	
#elif defined (SHADER_TEXTURE_EMISSIVE) || defined (SHADER_SNOW_NORMALMAP)
    
    varying_uv0         = (attrib_uv0 * DECOMPRESSION_MULTIPLIER_UV);

#else
	// Static uv's...
#if defined (SHADER_TEXTURE)
	varying_uv0			= (attrib_uv0 * DECOMPRESSION_MULTIPLIER_UV);
#if defined (SHADER_TEXTURE_MULTI_UV2)
	varying_uv1			= (attrib_uv1 * DECOMPRESSION_MULTIPLIER_UV);
#endif
#endif
    
#endif
	
#if defined (SHADER_TEXTURE_UVOFFSET)
	varying_uv0			+= uniform_uvOffset;
#endif
	
#endif
	
	//--------------------------------------------------------------------------
	// Map cloud shadows...
	//--------------------------------------------------------------------------

#if defined (SHADER_MAP_CLOUD_SHADOWS)
	mediump	vec2	cloudShadowUVOffset	= fract(uniform_animUV.xy * vec2(1.0/55.0, 1.0/75.0));
	mediump vec2	cloudShadowUV		= vec2(posWorld.x, posWorld.z) * (1.0 / 50.0);
	
	//TODO - This is a temporary offset to shift the cloud shadows, this should be adjustable or possible just linked to the light direction...
	cloudShadowUV += vec2(-0.1, 0.1);
	
	varying_cloud_uv0uv1.xy				= (cloudShadowUV - vec2(0.5, 0.5));
	varying_cloud_uv0uv1.zw				= (((cloudShadowUV * 2.0)) + cloudShadowUVOffset);
#endif
	
	//--------------------------------------------------------------------------
	// Lighting...
	//--------------------------------------------------------------------------
	
#if defined (SHADER_LIGHT) || defined (SHADER_VCOL) || defined (SHADER_FIXED_COLOUR)
	mediump vec4	output_rgba	= vec4(1.0, 1.0, 1.0, 1.0);
#endif
	
#if defined (SHADER_LIGHT)
	mediump vec3	lighting = vec3(0.0, 0.0, 0.0);
#endif
	
#if defined (SHADER_VCOL) || defined (SHADER_MARSH_WATER)
#if defined (SHADER_LIGHT) && defined (SHADER_VCOL_LIGHT_BAKE)
//	lighting		+= attrib_rgba.xyz;
//  lighting		+= pow(attrib_rgba.xyz, vec3(2.2));
	lighting		+= (attrib_rgba.xyz * LIGHT_BAKE_RGB_SCALE);
#if defined (SHADER_REFLECTIVE_SURFACE)
    output_rgba.w   = attrib_rgba.w;
#endif
#else
	output_rgba		*= attrib_rgba;
#endif
#endif
	
#if defined (SHADER_FIXED_COLOUR)
	output_rgba		*= uniform_colour;
#endif

	//--------------------------------------------------------------------------
	// World space vectors required by some of the lighting calculations...
#if (defined (SHADER_LIGHT) && defined (SHADER_LIGHT_SPECULAR) && !defined (SHADER_PER_PIXEL_SPECULAR)) || (defined (SHADER_LIGHT) && defined (SHADER_FOG) && defined (SHADER_FOG_SUN_HAZE)) || defined (SHADER_TEXTURE_REFLECTION_SPHERE)
	mediump vec3	eyeWorld	= normalize(posWorld.xyz - uniform_eyePosition);
#endif
	
//#if defined (SHADER_LIGHT)
//#if defined (SHADER_LIGHT_DIRECTIONAL) || defined (SHADER_LIGHT_CLASSIC)
//#if defined (SHADER_LIGHT_SPECULAR) //|| (defined (SHADER_FOG) && defined (SHADER_FOG_SUN_HAZE))
//#if defined (SHADER_LIGHT_DIRECTIONAL)
//    mediump vec3    lightWorld    = uniform_lightDirection;
//#elif defined (SHADER_LIGHT_CLASSIC)
//    mediump vec3    lightWorld    = vec3(uniform_lightDirectionMat[0].x, uniform_lightDirectionMat[1].x, uniform_lightDirectionMat[2].x);
//#endif
//#endif
//#endif
//#endif

#if defined (SHADER_LIGHT)
#if (defined (SHADER_LIGHT_SPECULAR) && !defined (SHADER_PER_PIXEL_SPECULAR) && !defined (SHADER_SPECULAR_NORMAL_UP)) || (defined (SHADER_FOG) && defined (SHADER_FOG_SUN_HAZE))
#if defined (SHADER_LIGHT_DIRECTIONAL)
    mediump vec3    lightWorld    = uniform_lightDirection;
#elif defined (SHADER_LIGHT_CLASSIC)
    mediump vec3    lightWorld    = vec3(uniform_lightDirectionMat[0].x, uniform_lightDirectionMat[1].x, uniform_lightDirectionMat[2].x);
#endif
#endif
#endif

	//--------------------------------------------------------------------------
	// Point light...
#if defined (SHADER_LIGHT) && defined (SHADER_LIGHT_POINT)
	lighting += calculatePointLight(posWorld.xyz, normalWorld, uniform_pointLightPosition, uniform_pointLightAttenuation, uniform_pointLightColour);
#endif

	//--------------------------------------------------------------------------
	// Spherical harmonic lighting...
#if defined (SHADER_LIGHT) && defined (SHADER_LIGHT_SH)
	lighting += calculateSphericalHarmonics(normalWorld);
#endif
	
	//--------------------------------------------------------------------------
	// Ambient lighting...
#if defined (SHADER_LIGHT) && defined (SHADER_LIGHT_AMBIENT)
	lighting += uniform_lightAmbient;
#endif
	
	//--------------------------------------------------------------------------
	// Old directional lighting (ambient + 3 directional lights)...
#if defined (SHADER_LIGHT) && defined (SHADER_LIGHT_CLASSIC)
	vec4	normalLightDots = calculateClassicLightingDots(vec4(normalWorld, 0.0), uniform_lightDirectionMat);
#if defined (SHADER_SNOW_DEFORMATION)
	normalLightDots.x *= (1.0 - (deformTexel.z * 0.2));
#endif
	lighting += (uniform_lightColourMat * normalLightDots).xyz;
#endif

	//--------------------------------------------------------------------------
	// Special 'constant' lighting, this is a variation on 'emissive' that actually uses the ambient and 1st directional light colours...
	// ...this is useful for things which do not require directional lighting but still need to change with the lighting setup...
#if !defined (SHADER_LIGHT) && defined (SHADER_LIGHT_CONSTANT)
	output_rgba.xyz = (uniform_lightAmbient + (output_rgba.xyz * uniform_lightColour));
#if !defined (SHADER_GAMMA_CORRECT)
	output_rgba.xyz = pow(output_rgba.xyz, vec3(1.0/2.2));
#endif
#endif
	
	//--------------------------------------------------------------------------
	// Diffuse lighting modifiers...
#if defined (SHADER_GRASS_DEFORMATION)
	// Darken the grass slightly when deformed...
	mediump float grassDarkenScale = min(1.0, (1.5 - (forceIntensity * 0.8)));
	grassDarkenScale = (grassDarkenScale * grassNormalModifier) + (1.0 - grassNormalModifier);
	lighting *= grassDarkenScale;
#endif
	
    //--------------------------------------------------------------------------
    // Directional lighting (part 1)...
#if defined (SHADER_LIGHT)
#if (defined (SHADER_LIGHT_DIRECTIONAL) && !defined (SHADER_SNOW)) || (defined (SHADER_SHADOW_RECEIVE) && !defined (SHADER_LIGHT_CLASSIC))
    mediump float normalLightDot = max(dot(uniform_lightDirection, normalWorld), 0.0);
#endif
#endif

	//--------------------------------------------------------------------------
	// Shadows...
#if defined (SHADER_LIGHT) && (defined (SHADER_LIGHT_DIRECTIONAL) || defined (SHADER_LIGHT_CLASSIC))
	#if defined (SHADER_SHADOW_RECEIVE) || defined (SHADER_MAP_CLOUD_SHADOWS) || defined (SHADER_SNOW) || defined (SHADER_TEXTURE_SHADOWMAP)
		#if defined (SHADER_LIGHT_CLASSIC)
			mediump vec3 shadowLighting = max((lighting - (uniform_lightColourMat[0].xyz * normalLightDots.x)), 0.0);
			#if !defined (SHADER_GAMMA_CORRECT)
				shadowLighting = pow(shadowLighting, vec3(1.0/2.2));
			#endif
        #else
			#if defined (SHADER_GAMMA_CORRECT)
				mediump vec3 shadowLighting = lighting;
			#else
				mediump vec3 shadowLighting = pow(lighting, vec3(1.0/2.2));
			#endif
		#endif
            varying_shadowAmbient = (output_rgba.xyz * shadowLighting);
    #endif

    // Modify the shadow intensity, this allows us to produce strong shadows without a bright directional light...
	#if defined (SHADER_SHADOW_RECEIVE)
        #if defined (SHADER_LIGHT_CLASSIC)
            varying_shadowAmbient *= (1.0 - (normalLightDots.x * uniform_shadowLightingModifier));
        #else
            varying_shadowAmbient *= (1.0 - (normalLightDot * uniform_shadowLightingModifier));
        #endif
    
        varying_shadowUVZW = (uniform_shadowProjectionMatrix * posWorld);
    #endif
    
//    //LM: Experimenting with normal offset for shadows, disabled for now but worth revisiting...
//    #if defined (SHADER_SHADOW_RECEIVE)
////#define SHADER_SHADOW_RECEIVE_NORMAL_OFFSET
////#define SHADER_SHADOW_RECEIVE_DEPTH_OFFSET
//    
//    mediump float normalOffsetScale = 0.0025;   //0.1;//0.05;
//    mediump float depthOffset = -0.0001;    //0.0025;//0.0002;//0.05;
//
//        #if defined (SHADER_SHADOW_RECEIVE_NORMAL_OFFSET) && defined (SHADER_LIGHT) && (defined (SHADER_LIGHT_DIRECTIONAL) || defined (SHADER_LIGHT_CLASSIC))
//    
//            highp vec4 shadowPosWorld = posWorld;
//            highp float cosLight = abs(dot(lightWorld.xyz, normalWorld.xyz));
//            highp float slopeScale = clamp((1.0 - cosLight), 0.0, 1.0);
//            normalOffsetScale *= slopeScale;
//            shadowPosWorld.xyz += (normalWorld.xyz * normalOffsetScale);
//    
//            varying_shadowUVZW = (uniform_shadowProjectionMatrix * shadowPosWorld);
//
//            #if defined (SHADER_SHADOW_RECEIVE_DEPTH_OFFSET)
//                //varying_shadowUVZW.z -= (slopeScale * depthOffset * varying_shadowUVZW.w);
//                varying_shadowUVZW.z -= (depthOffset * varying_shadowUVZW.w);
//                //varying_shadowUVZW.z -= (cosLight * depthOffset * varying_shadowUVZW.w);
//            #endif
//    
//        #else
//    
//            varying_shadowUVZW = (uniform_shadowProjectionMatrix * posWorld);
//
//            #if defined (SHADER_SHADOW_RECEIVE_DEPTH_OFFSET)
//                varying_shadowUVZW.z -= (depthOffset * varying_shadowUVZW.w);
//            #endif
//        #endif
//    #endif
    
#endif

    //--------------------------------------------------------------------------
    // Directional lighting (part 2)...
#if defined (SHADER_SNOW)
    lighting += uniform_lightColour;
#elif defined (SHADER_LIGHT) && defined (SHADER_LIGHT_DIRECTIONAL)
    lighting += (uniform_lightColour * normalLightDot);
#endif
    
	//--------------------------------------------------------------------------
	// Reflections...
#if defined (SHADER_TEXTURE_REFLECTION_SPHERE)
	varying_reflectionUV.xyz = reflect(eyeWorld, normalWorld);
#endif

#if defined (SHADER_TEXTURE_REFLECTION_LATLONG)
	mediump vec3	r = normalize(reflect(eyeWorld, normalWorld));
	r.x = -r.x;
	//varying_reflectionUV = vec4(r.xyz, 0.0);				// XZ and Y Per pixel...
	varying_reflectionUV = vec4(r.xyz, (acos(r.y) / PI));	// XZ per pixel and Y per vertex (y does not need to wrap)...
#endif
	
    //--------------------------------------------------------------------------
    // Refraction...
#if defined (SHADER_TEXTURE_REFRACTION)
    highp vec4 refractPosWorld = (posWorld - (vec4(normalWorld, 0.0) * 0.05));
    refractPosWorld = (uniform_viewProjectionMatrix * refractPosWorld);
    varying_refractionUV = ((refractPosWorld.xy / refractPosWorld.w) * uniform_refractUVScaleOffset.xy) + uniform_refractUVScaleOffset.zw;
#endif
    
	//--------------------------------------------------------------------------
	// Specular...
#if defined (SHADER_LIGHT) && defined (SHADER_LIGHT_SPECULAR)
//	// CORRECT: Full per-pixel specular...
//	varying_specLightVec	= lightWorld;
//	varying_specNormalVec	= normalWorld;
//	varying_specEyeVec		= eyeWorld;

//	// FASTER: Perform the reflect per vertex and interpolate (and normalize) per pixel, this actually looks identical...
//	varying_specRefVec	= reflect(lightWorld, normalWorld);
//	varying_specEyeVec	= eyeWorld;
	
#if defined (SHADER_LIGHT_DIRECTIONAL)
    mediump float specDot = normalLightDot;
#else
    mediump float specDot = normalLightDots.x;
#endif

//    mediump vec3	specLightDirWorld = lightWorld;
    mediump vec3	specLightDirWorld = normalize(uniform_specularDirection);

#if defined (SHADER_PER_PIXEL_SPECULAR)
    varying_specLightVec = vec4(specLightDirWorld, specDot);
    varying_specPosWorld = posWorld.xyz;
#else
	// MAYA: This is how the Maya shader fx shader does specular...
    varying_specRefVec = vec4(normalize(specLightDirWorld - eyeWorld), specDot);
#endif
                               
#if !defined (SHADER_SPECULAR_NORMAL_UP)
    varying_specNormal	= normalWorld;
#endif
#endif

	//--------------------------------------------------------------------------
	// Gamma correct the lighting result...

#if defined (SHADER_LIGHT)
#if !defined (SHADER_GAMMA_CORRECT)
	lighting = pow(lighting, vec3(1.0/2.2));
#endif
	output_rgba.xyz *= lighting;
#endif
	
	//--------------------------------------------------------------------------
	// Fog...
	//--------------------------------------------------------------------------

#if defined (SHADER_FOG)
	mediump float	fogValue 	= calculateFogValue(posWorld, uniform_fogPlaneY, uniform_fogPlaneZ, uniform_fogColour.w);
#if defined (SHADER_FADE_FOG_BY_ALPHA)
    fogValue *= attrib_rgba.w;
    output_rgba.w = 1.0;
#endif
#if defined (SHADER_FOG_SUN_HAZE) && defined (SHADER_LIGHT)
	varying_fogColour = calculateFogColourWithSunHaze(uniform_fogColour.xyz, fogValue, uniform_sunHazeColourFog, eyeWorld, uniform_horizonSunLightDirection);
#else
	varying_fogColour = calculateFogColour(uniform_fogColour.xyz, fogValue);
#endif
#endif
    
    //--------------------------------------------------------------------------
    // Dissolve effect...
    //--------------------------------------------------------------------------
    
#if defined (SHADER_DISSOLVE)
    // Adjust the animation value to so that the actual 0-1 range is clear of the line width...
    mediump float lineWidth = uniform_emissiveColour.w;
    mediump float lineMax = (uniform_animationValue * (1.0 + lineWidth));
    mediump float lineMin = (lineMax - lineWidth);
    
    varying_dissolve = vec2(lineMin, lineMax);
#endif

	//--------------------------------------------------------------------------
	// Final vertex colour...
#if defined (SHADER_LIGHT) || defined (SHADER_VCOL) || defined (SHADER_FIXED_COLOUR)
	varying_rgba	= output_rgba;
#endif

#if defined (SHADER_TEXTURE_MULTI_VRED)
    varying_rgba.w = attrib_rgba.x;
#endif
    
}

//------------------------------------------------------------------------------
