//
//  Shader_PreLitTexReflectiveWater.fsh
//  CityBuilder
//
//  Created by Lee Metcalfe on 11/06/2011.
//  Copyright (c) 2011 Midoki Ltd. All rights reserved.
//

//------------------------------------------------------------------------------

// Defines and macros to allow the shader code to be compiled for ES2 and ES3...
#if defined ES_300

#define ATTRIBUTE				in
#define VARYING_IN				in
#define VARYING_OUT				out
#define UNIFORM					uniform
#define FRAGCOLOUR				out_fragColour
#define SAMPLER2D				uniform sampler2D
#define SAMPLER2DSHADOW			uniform mediump sampler2DShadow
#define TEXTURE2D(a,b)			texture(a,b)
#define TEXTURE2DSHADOW(a,b)	texture(a,b.xyz/b.w)

#if defined (ANDROID)
#define TEXTURE2DBIAS(a,b,c)	texture(a,b)
#else
#define TEXTURE2DBIAS(a,b,c)	texture(a,b,c)
#endif

out lowp vec4 out_fragColour;

#else

#define ATTRIBUTE				attribute
#define VARYING_IN				varying
#define VARYING_OUT				varying
#define UNIFORM					uniform
#define FRAGCOLOUR				gl_FragColor
#define SAMPLER2D				uniform sampler2D
#define TEXTURE2D(a,b)			texture2D(a,b)

#if defined (OSX)
#define SAMPLER2DSHADOW			uniform sampler2DShadow
#define TEXTURE2DSHADOW(a,b)	shadow2DProj(a,b).x
#else
#define SAMPLER2DSHADOW			uniform sampler2D
#define TEXTURE2DSHADOW(a,b)	texture2DProj(a,b).x
#endif

#if defined (ANDROID)
#define TEXTURE2DBIAS(a,b,c)	texture2D(a,b)
#else
#define TEXTURE2DBIAS(a,b,c)	texture2D(a,b,c)
#endif

#endif

//------------------------------------------------------------------------------
// Some special cases...
//------------------------------------------------------------------------------

#if defined (SHADER_MARSH_WATER)
#undef SHADER_TEXTURE_NORMAL
#undef SHADER_TEXTURE_NORMAL2
#undef SHADER_TEXTURE_ANIMUV
#undef SHADER_TEXTURE_REFLECTION_PLANE

#define SHADER_TEXTURE_NORMAL
#define SHADER_TEXTURE_NORMAL2
#define SHADER_TEXTURE_ANIMUV
#define SHADER_TEXTURE_REFLECTION_PLANE
#endif

//------------------------------------------------------------------------------

#if defined (SHADER_LIGHT)
//#define SHADER_LIGHT_SH
#define SHADER_LIGHT_DIRECTIONAL
#endif

//------------------------------------------------------------------------------

#define REFLECTION_SCREEN_X_SCALE	(0.9)
#define REFLECTION_SCREEN_Y_SCALE	(0.8)

//------------------------------------------------------------------------------

UNIFORM highp	mat4	uniform_viewProjectionMatrix;

#if defined (SHADER_LIGHT_DIRECTIONAL)
UNIFORM mediump vec3	uniform_lightDirection;
UNIFORM highp	vec3    uniform_lightColour;
#endif

UNIFORM highp	vec3	uniform_eyePosition;

#if defined (SHADER_LIGHT_SPECULAR)
UNIFORM mediump vec4    uniform_specularRGBPow;
//UNIFORM mediump vec3	uniform_horizonSunLightDirection;
UNIFORM mediump vec3	uniform_specularDirection;
#endif

UNIFORM highp	vec4	uniform_oceanWaterSettings;

//------------------------------------------------------------------------------

SAMPLER2D   uniform_texture0;

SAMPLER2D	uniform_refractionmap;

#if defined (SHADER_TEXTURE_NORMAL)
SAMPLER2D	uniform_normalmap;
//SAMPLER2D	uniform_normalmap2;
SAMPLER2D	uniform_alphamap;
#endif

#if defined (SHADER_SHADOW_RECEIVE)
SAMPLER2DSHADOW uniform_shadowmap;
#endif

SAMPLER2D	uniform_ambientmap;

//------------------------------------------------------------------------------

VARYING_IN highp	vec4	varying_posWorld;

VARYING_IN lowp     vec4	varying_rgba_vcol;
VARYING_IN mediump  vec3	varying_rgba_light;

#if defined (SHADER_TEXTURE_NORMAL)
VARYING_IN lowp     vec2	varying_uv0;
VARYING_IN mediump	vec2	varying_uv1;
VARYING_IN mediump	vec2	varying_uv2;
#else
VARYING_IN highp	vec3	varying_uvw;
VARYING_IN lowp     vec2	varying_uv2;
#endif

#if defined (SHADER_FOG)
VARYING_IN mediump	vec4 varying_fogColour;
#endif

#if defined (SHADER_SHADOW_RECEIVE)
VARYING_IN highp	vec4	varying_shadowUVZW;
#endif

//------------------------------------------------------------------------------

void main()
{
	mediump vec3	mask = TEXTURE2D(uniform_ambientmap, varying_uv0.xy).xyz;
    mediump float   normalHeight = varying_rgba_vcol.w;// * mask.x;
    mediump float   reflectionLevel = mask.x;//(1.0f - mask.x);//(1.0 - normalHeight);

    normalHeight    *= uniform_oceanWaterSettings.z;
    reflectionLevel *= uniform_oceanWaterSettings.w;

	mediump vec3	lightDirWorld = normalize(uniform_lightDirection);

    mediump vec4	texel = TEXTURE2DBIAS(uniform_texture0, varying_uv0,  -1.0) * varying_rgba_vcol;
	
	// Adjust the amount of normal map distortion based on distance, without this the ripples tend to get lost far from the camera...
	mediump vec3	eyeToPos = (varying_posWorld.xyz - uniform_eyePosition.xyz);
	mediump vec3	eyeDirWorld = normalize(eyeToPos);

#if defined (SHADER_TEXTURE_NORMAL)
	mediump float	eyeToPosDist = length(eyeToPos);
	mediump float	distortionScale = (clamp(eyeToPosDist, 4.0, 16.0) / 4.0);

	// METHOD 2: Perform a proper reflection calculation, this can be manipulated with a normal map...
	mediump vec3	normalWorld0 = (TEXTURE2D(uniform_normalmap, varying_uv1.xy).xyz * 2.0) - 1.0;
	mediump vec3	normalWorld1 = (TEXTURE2D(uniform_alphamap, varying_uv2.xy).xyz * 2.0) - 1.0;
//	mediump vec3	normalWorldA = normalize((normalWorld0 * 0.5) + (normalWorld1 * 0.5));
    mediump vec3	normalWorldA = (normalWorld0 + normalWorld1);

	// Correct the 'space' of the normal map...
	mediump vec3	normalWorld = vec3(normalWorldA.x, normalWorldA.z, -normalWorldA.y);

    // Apply the normal height to the normal...
    normalWorld.xz *= normalHeight;
    normalWorld = normalize(normalWorld);
    
//	mediump vec3	reflectWorld = normalize(reflect(eyeDirWorld, normalWorld)) * vec3(0.1, -1.0, 0.1) * distortionScale;
//	mediump vec3	reflectWorld = normalize(reflect(eyeDirWorld, normalWorld)) * vec3(0.01, -1.0, 0.01) * distortionScale;
	mediump vec3	reflectWorld = normalize(reflect(eyeDirWorld, normalWorld)) * vec3(0.1, -1.0, 0.1) * distortionScale;
	
    reflectWorld *= normalHeight;//mask.x;
	
	// Prevent the reflection vector causing the reflection map to be sampled above the level of the reflection plane...
	reflectWorld.y = max(0.0, reflectWorld.y);
	mediump vec4	posReflected = (varying_posWorld - vec4(reflectWorld, 0.0));
	mediump vec4	posReflectedScreen = (uniform_viewProjectionMatrix * posReflected);
	mediump vec2	uv = ((posReflectedScreen.xy / posReflectedScreen.w) * vec2((0.5 * REFLECTION_SCREEN_X_SCALE), (0.5 * REFLECTION_SCREEN_Y_SCALE))) + 0.5;
#else
	mediump vec2	uv = ((varying_uvw.xy / varying_uvw.z) * vec2((0.5 * REFLECTION_SCREEN_X_SCALE), (0.5 * REFLECTION_SCREEN_Y_SCALE))) + 0.5;
	mediump vec3	normalWorld = vec3(0.0, 1.0, 0.0);
#endif
	
	mediump vec3	reflectionTexel = TEXTURE2D(uniform_refractionmap, uv).xyz;
//    mediump vec3	reflectionTexel = vec3(1.0, 0.5, 1.0);

    //-----------------------
    // Diffuse lighting...
#if defined (SHADER_LIGHT)
    mediump float	normalLightDot = max(0.0, dot(lightDirWorld, normalWorld));
    
#if defined (SHADER_SHADOW_RECEIVE)
    mediump float   shadow = TEXTURE2DSHADOW(uniform_shadowmap, varying_shadowUVZW);
    //light = (light * shadow) + (0.7 * (1.0 - shadow));
    normalLightDot *= shadow;
#endif
    
    mediump vec3 lighting = (uniform_lightColour.xyz * normalLightDot);
    
    lighting += varying_rgba_light;
    
#if !defined (SHADER_GAMMA_CORRECT)
    lighting.xyz = pow(lighting.xyz, vec3(1.0/2.2));
#endif
    
    texel.xyz *= lighting.xyz;
#endif
    
    
//	texel.xyz *= reflectionTexel;
//	texel.xyz = ((texel.xyz * reflectionTexel) * mask) + (texel.xyz * (1.0 - mask));
    texel.xyz = (reflectionTexel * reflectionLevel) + (texel.xyz * (1.0 - reflectionLevel));

    //------------------------
    // Specular lighting...
#if defined (SHADER_LIGHT)
//	texel *= light;
//	texel.xyz = ((texel.xyz * light) * mask) + (texel.xyz * (1.0 - mask));
//  texel.xyz = ((texel.xyz * normalLightDot) * normalHeight) + (texel.xyz * (1.0 - normalHeight));

#if defined (SHADER_LIGHT_SPECULAR)
	// Specular...
//  mediump vec3	specLightDirWorld = normalize(uniform_horizonSunLightDirection);
    mediump vec3	specLightDirWorld = normalize(uniform_specularDirection);
//	mediump float	specDot = max(0.0, dot(eyeDirWorld, normalize(reflect(specLightDirWorld, normalWorld))));
//	mediump	float	specCoeff = pow(specDot, uniform_specularRGBPow.w) * 0.5;
    mediump	vec3    specRefVec = normalize(specLightDirWorld - eyeDirWorld);
    mediump	float	specCoeff = max(0.0, dot(specRefVec, normalWorld));
    specCoeff = (pow(specCoeff, uniform_specularRGBPow.w) * normalLightDot);
#if defined (SHADER_SHADOW_RECEIVE)
	specCoeff *= shadow;
#endif
	texel.xyz += (uniform_specularRGBPow.xyz * specCoeff);
#endif
#endif
	
#if defined (SHADER_FOG)
	texel.xyz = (texel.xyz * varying_fogColour.w) + varying_fogColour.xyz;
#endif

	FRAGCOLOUR = vec4(texel.xyz, 1.0);
}

//------------------------------------------------------------------------------
