//
//  Shader_PreLitTexReflectiveWater.vsh
//  CityBuilder
//
//  Created by Lee Metcalfe on 11/06/2011.
//  Copyright (c) 2011 Midoki Ltd. All rights reserved.
//

//------------------------------------------------------------------------------

// Defines and macros to allow the shader code to be compiled for ES2 and ES3...
#if defined ES_300

#define ATTRIBUTE				in
#define VARYING_IN				in
#define VARYING_OUT				out
#define UNIFORM					uniform
#define FRAGCOLOUR				out_fragColour
#define SAMPLER2D				uniform sampler2D
#define SAMPLER2DSHADOW			uniform mediump sampler2DShadow
#define TEXTURE2D(a,b)			texture(a,b)
#define TEXTURE2DSHADOW(a,b)	texture(a,b.xyz/b.w)

#if defined (ANDROID)
#define TEXTURE2DBIAS(a,b,c)	texture(a,b)
#else
#define TEXTURE2DBIAS(a,b,c)	texture(a,b,c)
#endif

#else

#define ATTRIBUTE				attribute
#define VARYING_IN				varying
#define VARYING_OUT				varying
#define UNIFORM					uniform
#define FRAGCOLOUR				gl_FragColor
#define SAMPLER2D				uniform sampler2D
#define TEXTURE2D(a,b)			texture2D(a,b)

#if defined (OSX)
#define SAMPLER2DSHADOW			uniform sampler2DShadow
#define TEXTURE2DSHADOW(a,b)	shadow2DProj(a,b).x
#else
#define SAMPLER2DSHADOW			uniform sampler2D
#define TEXTURE2DSHADOW(a,b)	texture2DProj(a,b).x
#endif

#if defined (ANDROID)
#define TEXTURE2DBIAS(a,b,c)	texture2D(a,b)
#else
#define TEXTURE2DBIAS(a,b,c)	texture2D(a,b,c)
#endif

#endif

//------------------------------------------------------------------------------
// Some special cases...
//------------------------------------------------------------------------------

#if defined (SHADER_MARSH_WATER)
#undef SHADER_TEXTURE_NORMAL
#undef SHADER_TEXTURE_NORMAL2
#undef SHADER_TEXTURE_ANIMUV
#undef SHADER_TEXTURE_REFLECTION_PLANE

#define SHADER_TEXTURE_NORMAL
#define SHADER_TEXTURE_NORMAL2
#define SHADER_TEXTURE_ANIMUV
#define SHADER_TEXTURE_REFLECTION_PLANE
#endif

//------------------------------------------------------------------------------

#if defined (SHADER_LIGHT)
#define SHADER_LIGHT_SH
//#define SHADER_LIGHT_DIRECTIONAL
#endif

//------------------------------------------------------------------------------

#define DECOMPRESSION_MULTIPLIER_UV				(1.0 / 1024.0)
#define DECOMPRESSION_MULTIPLIER_WEIGHT			(1.0 / 255.0)

//------------------------------------------------------------------------------

ATTRIBUTE vec4 attrib_position;

#if defined (SHADER_LIGHT)
ATTRIBUTE vec3 attrib_normal;
#endif

ATTRIBUTE vec2 attrib_uv0;
ATTRIBUTE vec2 attrib_uv1;

ATTRIBUTE vec4 attrib_rgba;

//------------------------------------------------------------------------------

#if defined (SHADER_TEXTURE_NORMAL)
UNIFORM highp	mat4	uniform_viewProjectionMatrix;
#else
UNIFORM highp	mat4	uniform_modelViewProjectionMatrix;
#endif

UNIFORM highp	mat4	uniform_modelMatrix;

UNIFORM highp	vec3	uniform_eyePosition;

UNIFORM highp	vec4	uniform_animUV;
UNIFORM highp	vec4	uniform_oceanWaterSettings;

#if defined (SHADER_FOG)
#define SHADER_FOG_EXP2
UNIFORM mediump	vec4	uniform_fogPlaneY;
UNIFORM mediump	vec4	uniform_fogPlaneZ;
UNIFORM mediump vec4	uniform_fogColour;
#endif

#if defined (SHADER_SHADOW_RECEIVE)
UNIFORM highp	mat4	uniform_shadowProjectionMatrix;
#endif

#if defined (SHADER_FOG_SUN_HAZE)
UNIFORM highp	vec4 uniform_sunHazeColourFog;
#endif

#if defined (SHADER_LIGHT)
UNIFORM mediump	vec3 uniform_lightDirection;
#endif

//------------------------------------------------------------------------------

#if defined (SHADER_LIGHT_SH)
UNIFORM mediump	vec4 uniform_SHCoeff[7];
#endif

//------------------------------------------------------------------------------

VARYING_OUT highp	vec4	varying_posWorld;

VARYING_OUT lowp     vec4	varying_rgba_vcol;
VARYING_OUT mediump vec3	varying_rgba_light;

#if defined (SHADER_TEXTURE_NORMAL)
VARYING_OUT lowp	vec2	varying_uv0;
VARYING_OUT mediump	vec2	varying_uv1;
VARYING_OUT mediump	vec2	varying_uv2;
#else
VARYING_OUT highp	vec3	varying_uvw;
VARYING_OUT lowp	vec2	varying_uv2;
#endif

#if defined (SHADER_FOG)
VARYING_OUT mediump	vec4 varying_fogColour;
#endif

#if defined (SHADER_SHADOW_RECEIVE)
VARYING_OUT highp	vec4	varying_shadowUVZW;
#endif


//------------------------------------------------------------------------------

#if defined (SHADER_LIGHT_SH)
mediump vec3 calculateSphericalHarmonics(mediump vec3 n)
{
    //	const mediump float C1 = 0.429043;
    //	const mediump float C2 = 0.511664;
    //	const mediump float C3 = 0.743125;
    //	const mediump float C4 = 0.886227;
    //	const mediump float C5 = 0.247708;
    
    mediump vec3	flip	= (n.xzy * n.zyx);
    mediump vec3	square	= (n * n);
    
    //	mediump vec3	result  = C4 * uniform_SHCoeff[0].xyz;
    //					result += C2 * 2.0 * uniform_SHCoeff[1].xyz * n.x;
    //					result += C2 * 2.0 * uniform_SHCoeff[2].xyz * n.y;
    //					result += C2 * 2.0 * uniform_SHCoeff[3].xyz * n.z;
    //					result += C1 * 2.0 * uniform_SHCoeff[4].xyz * flip.x;
    //					result += C1 * 2.0 * uniform_SHCoeff[5].xyz * flip.y;
    //					result += C1 * 2.0 * uniform_SHCoeff[6].xyz * flip.z;
    //					result += C3 * uniform_SHCoeff[7].xyz * (square.z - C5);
    //					result += C1 * uniform_SHCoeff[8].xyz * (square.x - square.y);
    
    //	// We now roll most of the constants above into the coefficient uniforms...
    //	mediump vec3	result  = uniform_SHCoeff[0].xyz;
    //					result += uniform_SHCoeff[1].xyz * n.x;
    //					result += uniform_SHCoeff[2].xyz * n.y;
    //					result += uniform_SHCoeff[3].xyz * n.z;
    //					result += uniform_SHCoeff[4].xyz * flip.x;
    //					result += uniform_SHCoeff[5].xyz * flip.y;
    //					result += uniform_SHCoeff[6].xyz * flip.z;
    //					result += uniform_SHCoeff[7].xyz * (square.z - C5);
    //					result += uniform_SHCoeff[8].xyz * (square.x - square.y);
    
    // We now roll most of the constants above into the coefficient uniforms, also now compressed into fewer vectors...
    mediump vec3	result  = uniform_SHCoeff[0].xyz;
    result += uniform_SHCoeff[1].xyz * n.x;
    result += uniform_SHCoeff[2].xyz * n.y;
    result += uniform_SHCoeff[3].xyz * n.z;
    result += uniform_SHCoeff[4].xyz * flip.x;
    result += uniform_SHCoeff[5].xyz * flip.y;
    result += uniform_SHCoeff[6].xyz * flip.z;
    result += vec3(uniform_SHCoeff[0].w, uniform_SHCoeff[1].w, uniform_SHCoeff[2].w) * (square.z - uniform_SHCoeff[6].w);
    result += vec3(uniform_SHCoeff[3].w, uniform_SHCoeff[4].w, uniform_SHCoeff[5].w) * (square.x - square.y);
    
    result = max(result, 0.0);
    
    return result;
}
#endif

//------------------------------------------------------------------------------

#if defined (SHADER_FOG)
mediump float calculateFogValue(mediump vec4 posModel, mediump vec4 fogPlaneY, mediump vec4 fogPlaneZ, mediump float fogDensity)
{
	mediump float	fogValue	= max(dot(fogPlaneY, posModel), dot(fogPlaneZ, posModel));
	fogValue	= max(fogValue, 0.0);
	
#if defined (SHADER_FOG_EXP2)
#define LOG2 1.442695
	mediump float	result		= exp2(-(fogValue*fogValue*LOG2));
	result		= clamp(result, 0.0, 1.0);
	result		= (1.0 - result) * fogDensity;
#else
	mediump float	result		= min(fogValue, 1.0) * fogDensity;
#endif
	
	return result;
}
#endif

//------------------------------------------------------------------------------

#if defined (SHADER_FOG)
mediump vec4 calculateFogColour(mediump vec3 fogColour, mediump float fogValue)
{
	mediump vec4	result = vec4((uniform_fogColour.xyz * fogValue), (1.0 - fogValue));
	
	return result;
}
#endif

//------------------------------------------------------------------------------

#if defined (SHADER_FOG) && defined (SHADER_FOG_SUN_HAZE) && defined (SHADER_LIGHT)
mediump vec4 calculateFogColourWithSunHaze(mediump vec3 fogColour, mediump float fogValue, mediump vec4 sunHazeFogColour, mediump vec3 eye, mediump vec3 light)
{
	mediump float	hazeFactor	= max(0.0, dot(light, eye));
	hazeFactor	= (hazeFactor * hazeFactor) * sunHazeFogColour.w;//pow(hazeFactor, 8.0);
	mediump vec3	hazeColour	= (sunHazeFogColour.xyz * hazeFactor) + (fogColour.xyz * (1.0 - hazeFactor));
	
	mediump vec4	result		= vec4((hazeColour.xyz * fogValue), (1.0 - fogValue));
	
	return result;
}
#endif

//------------------------------------------------------------------------------

void main()
{
    // Transform the position into screen space.
 //   vec3 posScaled = (attrib_position.xyz * DECOMPRESSION_MULTIPLIER_POSITION);
	vec3 posScaled	= (attrib_position.xyz * (attrib_position.w * (1.0 / 32767.0)));
	vec4 posScaled4 = vec4(posScaled, 1.0);
	
	highp vec4	posWorld = (uniform_modelMatrix * posScaled4);

	varying_posWorld = posWorld;

//#if defined (SHADER_TEXTURE_NORMAL)
//	highp vec4 posScreen = (uniform_viewProjectionMatrix * posWorld);
//
//	mediump	vec2 baseUVOffset0 = fract(uniform_animUV.xy * vec2(-0.036, -0.021) * 4.0);
////	varying_uv0 = (attrib_uv0 * DECOMPRESSION_MULTIPLIER_UV);
////	varying_uv0 = (posWorld.xz * 0.50) + baseUVOffset0;
//	varying_uv0 = (posWorld.xz * 0.75) + baseUVOffset0;
////	varying_uv0 = (posWorld.xz * 1.5) + baseUVOffset0;
//
//	mediump	vec2 baseUVOffset1 = fract(uniform_animUV.xy * vec2(-0.065, -0.035) * 2.5);
////	varying_uv1 = (posWorld.xz * 0.50) + baseUVOffset1;
//	varying_uv1 = (posWorld.xz * 0.25) + baseUVOffset1;
////	varying_uv1 = (posWorld.xz * 0.50) + baseUVOffset1;
//	
//	mediump	vec2 baseUVOffset2 = fract(uniform_animUV.xy * vec2(-0.036, -0.021) * 1.0);
//	varying_uv2 = (attrib_uv0 * DECOMPRESSION_MULTIPLIER_UV) + baseUVOffset2;
//#else
//	highp vec4 posScreen = (uniform_modelViewProjectionMatrix * posScaled4);
//	varying_uvw.xyz = posScreen.xyw;
//	
//	varying_uv2 = (attrib_uv0 * DECOMPRESSION_MULTIPLIER_UV);
//#endif
    
#if defined (SHADER_TEXTURE_NORMAL)
    highp vec4 posScreen = (uniform_viewProjectionMatrix * posWorld);
    
    mediump	vec2 baseUVOffset1 = fract(uniform_animUV.xy * uniform_animUV.zw);
    mediump	vec2 baseUVOffset2 = fract(uniform_animUV.xy * uniform_oceanWaterSettings.xy);

    varying_uv0 = (attrib_uv0 * DECOMPRESSION_MULTIPLIER_UV);
    varying_uv1 = (attrib_uv1 * DECOMPRESSION_MULTIPLIER_UV) + baseUVOffset1;
    varying_uv2 = (attrib_uv1 * DECOMPRESSION_MULTIPLIER_UV) + baseUVOffset2;
#else
    highp vec4 posScreen = (uniform_modelViewProjectionMatrix * posScaled4);
    varying_uvw.xyz = posScreen.xyw;
    
    varying_uv2 = (attrib_uv0 * DECOMPRESSION_MULTIPLIER_UV);
#endif

	gl_Position = posScreen;

	varying_rgba_vcol = attrib_rgba;

    //--------------------------------------------------------------------------
    // Spherical harmonic lighting...
    mediump vec3 normalWorld = attrib_normal;
#if defined (SHADER_LIGHT) && defined (SHADER_LIGHT_SH)
    mediump	vec3 lighting = calculateSphericalHarmonics(normalWorld);
//#if !defined (SHADER_GAMMA_CORRECT)
//    lighting = pow(lighting, vec3(1.0/2.2));
//#endif
    varying_rgba_light = lighting;
#endif

	//--------------------------------------------------------------------------
	// Fog...
	//--------------------------------------------------------------------------
	
#if defined (SHADER_FOG)
	mediump float	fogValue 	= calculateFogValue(posWorld, uniform_fogPlaneY, uniform_fogPlaneZ, uniform_fogColour.w);
	
#if defined (SHADER_FOG_SUN_HAZE) && defined (SHADER_LIGHT)
    mediump vec3	eyeWorld	= normalize(posWorld.xyz - uniform_eyePosition);
    mediump vec3	lightWorld	= uniform_lightDirection;
	varying_fogColour = calculateFogColourWithSunHaze(uniform_fogColour.xyz, fogValue, uniform_sunHazeColourFog, eyeWorld, lightWorld);
#else
	varying_fogColour = calculateFogColour(uniform_fogColour.xyz, fogValue);
#endif
#endif

//#if defined (SHADER_FOG)
//	float fogValueZ = max(dot(uniform_fogPlaneY, posScaled4), dot(uniform_fogPlaneZ, posScaled4));
//	fogValueZ = max(fogValueZ, 0.0);
//	float fogDensity = uniform_fogColour.w;
//#if defined (SHADER_FOG_EXP2)
//#define LOG2 1.442695
//	float fogResult = (1.0 - clamp(exp2(-(fogValueZ*fogValueZ*LOG2)), 0.0, 1.0)) * fogDensity;
//#else
//	float fogResult = (min(fogValueZ, 1.0) * fogDensity);
//#endif
//	
//	mediump vec3 fogColour = uniform_fogColour.xyz;
//	varying_fogColour = vec4((fogColour.xyz * fogResult), (1.0 - fogResult));
//#endif
	
#if defined (SHADER_SHADOW_RECEIVE)
	varying_shadowUVZW = (uniform_shadowProjectionMatrix * posWorld);
#endif
}

//------------------------------------------------------------------------------
